# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Generate the whole folw content."""

from typing import Any

from ..core import (
    ContentGenerator,
    ExporterContentError,
    ExporterContext,
    ExportPosition,
    ExportResult,
    PositionedContent,
    get_comment,
)
from .execution_generator import ExecutionGenerator
from .utils.common import generate_header


class FileGenerator(ContentGenerator):
    """Generate the complete flow notebook content."""

    def __init__(
        self,
        context: ExporterContext,
    ) -> None:
        """Initialize the notebook generator.

        Parameters
        ----------
        context : ExporterContext
            The exporter context containing dependencies.
        """
        self.context = context
        self.config = context.get_config()

    # pylint: disable=too-many-locals
    def generate(
        self,
        merged_result: ExportResult,
        is_async: bool,
        after_run: str,
        **kwargs: Any,
    ) -> str:
        """Generate the complete flow notebook content.

        Parameters
        ----------
        merged_result : ExportResult
            The merged export result containing all content.
        is_async : bool
            Whether to generate async conten
        after_run : str
            Additional content to add after the main flow execution.
        **kwargs : Any
            Additional keyword arguments for the generator.

        Returns
        -------
        str
            The complete flow notebook content.

        Raises
        ------
        ExporterContentError
            If there is no content to export.
        """
        # 1. Generate header
        header = self.get_header(merged_result)

        # 2. Generate imports
        imports_section = merged_result.get_content_by_position(
            ExportPosition.IMPORTS
        )

        # 3. Generate content sections
        tools_section = merged_result.get_content_by_position(
            ExportPosition.TOOLS
        )
        models_section = merged_result.get_content_by_position(
            ExportPosition.MODELS
        )
        agents_section = merged_result.get_content_by_position(
            ExportPosition.AGENTS,
            # Skip agent arguments (should already be there)
            skip_agent_arguments=True,
        )
        chats_content = merged_result.get_content_by_position(
            ExportPosition.CHATS
        )
        if not chats_content:
            raise ExporterContentError(
                "No content to export. Please ensure that the flow has chats."
            )
        after_chats = merged_result.get_content_by_position(
            ExportPosition.BOTTOM
        )

        main, call_main, execution_block = self._get_execution_content(
            chats_content=chats_content,
            is_async=is_async,
            after_run=after_run,
            for_notebook=self.config.for_notebook,
        )

        # 5. Combine everything
        everything: list[str] = [header]
        if imports_section:
            comment = get_comment(
                "Imports",
                for_notebook=self.config.for_notebook,
            )
            everything.append(comment)
            everything.append(
                "\n".join([entry.content for entry in imports_section])
            )
        if tools_section:
            comment = get_comment(
                "Tools",
                for_notebook=self.config.for_notebook,
            )
            everything.append(comment)
            everything.append(
                "\n".join([entry.content for entry in tools_section]) + "\n"
            )
        if models_section:
            comment = get_comment(
                "Models",
                for_notebook=self.config.for_notebook,
            )
            everything.append(comment)
            everything.append(
                "\n\n".join([entry.content for entry in models_section]) + "\n"
            )
        if agents_section:
            comment = get_comment(
                "Agents",
                for_notebook=self.config.for_notebook,
            )
            if self.config.for_notebook:
                comment += "# pyright: reportUnnecessaryIsInstance=false\n"
            everything.append(comment)
            everything.append(
                "\n\n".join([entry.content for entry in agents_section]) + "\n"
            )
        everything.append(main)
        if after_chats:
            everything.append(
                "\n".join([entry.content for entry in after_chats])
            )
        everything.append(call_main)
        if execution_block:
            everything.append(execution_block)

        return "\n".join(everything)

    def _get_execution_content(
        self,
        chats_content: list[PositionedContent],
        is_async: bool,
        for_notebook: bool,
        after_run: str,
    ) -> tuple[str, str, str]:
        cache_seed = (
            self.context.config.cache_seed if self.context.config else None
        )
        execution_gen = ExecutionGenerator()
        chat_contents = "\n".join(chat.content for chat in chats_content)
        main = execution_gen.generate_main_function(
            content=chat_contents,
            is_async=is_async,
            for_notebook=for_notebook,
            cache_seed=cache_seed,
            after_run=after_run,
        )
        call_main = execution_gen.generate_call_main_function(
            is_async=is_async,
            for_notebook=for_notebook,
        )
        execution_block = (
            execution_gen.generate_execution_block(
                is_async=is_async,
            )
            if not for_notebook
            else ""
        )
        return main, call_main, execution_block

    def get_header(self, merged_result: ExportResult) -> str:
        """Get or generate the header for the script.

        Parameters
        ----------
        merged_result : ExportResult
            The merged export result containing all content.

        Returns
        -------
        str
            The header content.
        """
        from_result = merged_result.get_content_by_position(ExportPosition.TOP)
        if not from_result:
            return generate_header(
                name=self.config.name,
                description=self.config.description,
                requirements=self.config.requirements,
                tags=self.config.tags,
                for_notebook=self.config.for_notebook,
            )
        header_string = "\n".join(content.content for content in from_result)
        while not header_string.endswith("\n\n"):
            header_string += "\n"
        return header_string



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Export tool."""

from .exporter import ToolsExporter
from .factory import create_tools_exporter

__all__ = [
    "ToolsExporter",
    "create_tools_exporter",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utilities for running code."""

import asyncio
import datetime
import io
import os
import re
import shutil
import subprocess
import sys
import tempfile
from contextlib import asynccontextmanager, contextmanager
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Iterator,
    Optional,
    Set,
    Union,
)

from .environment import in_virtualenv, is_root
from .gen_seq_diagram import generate_sequence_diagram

# pylint: disable=import-outside-toplevel


@contextmanager
def chdir(to: Union[str, Path]) -> Iterator[None]:
    """Change the current working directory in a context.

    Parameters
    ----------
    to : Union[str, Path]
        The directory to change to.

    Yields
    ------
    Iterator[None]
        The context manager.
    """
    old_cwd = str(os.getcwd())
    os.chdir(to)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@asynccontextmanager
async def a_chdir(to: Union[str, Path]) -> AsyncIterator[None]:
    """Asynchronously change the current working directory in a context.

    Parameters
    ----------
    to : Union[str, Path]
        The directory to change to.

    Yields
    ------
    AsyncIterator[None]
        The async context manager.
    """
    old_cwd = str(os.getcwd())
    os.chdir(to)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def strip_ansi(text: str) -> str:
    """Remove ANSI escape sequences from text.

    Parameters
    ----------
    text : str
        The text to strip.

    Returns
    -------
    str
        The text without ANSI escape sequences.
    """
    ansi_pattern = re.compile(r"\x1b\[[0-9;]*m|\x1b\[.*?[@-~]")
    return ansi_pattern.sub("", text)


def before_run(
    output_path: Optional[Union[str, Path]],
    uploads_root: Optional[Union[str, Path]],
) -> str:
    """Actions to perform before running the flow.

    Parameters
    ----------
    output_path : Optional[Union[str, Path]]
        The output path.
    uploads_root : Optional[Union[str, Path]]
        The runtime uploads root.

    Returns
    -------
    str
        The file name.
    """
    if not uploads_root:
        uploads_root = Path(tempfile.mkdtemp())
    else:
        uploads_root = Path(uploads_root)
    if not uploads_root.exists():
        uploads_root.mkdir(parents=True)
    file_name = "waldiez_flow.py" if not output_path else Path(output_path).name
    if file_name.endswith((".json", ".waldiez")):
        file_name = file_name.replace(".json", ".py").replace(".waldiez", ".py")
    if not file_name.endswith(".py"):
        file_name += ".py"
    return file_name


def install_requirements(
    extra_requirements: Set[str], printer: Callable[..., None]
) -> None:
    """Install the requirements.

    Parameters
    ----------
    extra_requirements : Set[str]
        The extra requirements.
    printer : Callable[..., None]
        The printer function.
    """
    requirements_string = ", ".join(extra_requirements)
    printer(f"Installing requirements: {requirements_string}")
    pip_install = [sys.executable, "-m", "pip", "install"]
    break_system_packages = ""
    if not in_virtualenv():  # it should
        # if not, let's try to install as user
        # not sure if --break-system-packages is safe,
        # but it might fail if we don't
        break_system_packages = os.environ.get("PIP_BREAK_SYSTEM_PACKAGES", "")
        os.environ["PIP_BREAK_SYSTEM_PACKAGES"] = "1"
        if not is_root():
            pip_install.append("--user")
    pip_install.extend(extra_requirements)
    # pylint: disable=too-many-try-statements
    try:
        with subprocess.Popen(
            pip_install,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            if proc.stdout:
                for line in io.TextIOWrapper(proc.stdout, encoding="utf-8"):
                    printer(strip_ansi(line.strip()))
            if proc.stderr:
                for line in io.TextIOWrapper(proc.stderr, encoding="utf-8"):
                    printer(strip_ansi(line.strip()))
    finally:
        if not in_virtualenv():
            # restore the old env var
            if break_system_packages:
                os.environ["PIP_BREAK_SYSTEM_PACKAGES"] = break_system_packages
            else:
                del os.environ["PIP_BREAK_SYSTEM_PACKAGES"]


async def a_install_requirements(
    extra_requirements: Set[str], printer: Callable[..., None]
) -> None:
    """Install the requirements asynchronously.

    Parameters
    ----------
    extra_requirements : Set[str]
        The extra requirements.
    printer : Callable[..., None]
        The printer function.
    """
    requirements_string = ", ".join(extra_requirements)
    printer(f"Installing requirements: {requirements_string}")
    pip_install = [sys.executable, "-m", "pip", "install"]
    break_system_packages = ""
    if not in_virtualenv():
        break_system_packages = os.environ.get("PIP_BREAK_SYSTEM_PACKAGES", "")
        os.environ["PIP_BREAK_SYSTEM_PACKAGES"] = "1"
        if not is_root():
            pip_install.extend(["--user"])
    pip_install.extend(extra_requirements)
    # pylint: disable=too-many-try-statements
    try:
        proc = await asyncio.create_subprocess_exec(
            *pip_install,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        if proc.stdout:
            async for line in proc.stdout:
                printer(strip_ansi(line.decode().strip()))
        if proc.stderr:
            async for line in proc.stderr:
                printer(strip_ansi(line.decode().strip()))
    finally:
        if not in_virtualenv():
            if break_system_packages:
                os.environ["PIP_BREAK_SYSTEM_PACKAGES"] = break_system_packages
            else:
                del os.environ["PIP_BREAK_SYSTEM_PACKAGES"]


def after_run(
    temp_dir: Path,
    output_path: Optional[Union[str, Path]],
    printer: Callable[..., None],
    flow_name: str,
    skip_mmd: bool = False,
) -> None:
    """Actions to perform after running the flow.

    Parameters
    ----------
    temp_dir : Path
        The temporary directory.
    output_path : Optional[Union[str, Path]]
        The output path.
    printer : Callable[..., None]
        The printer function.
    flow_name : str
        The flow name.
    skip_mmd : bool, optional
        Whether to skip the mermaid sequence diagram generation,
        by default, False
    """
    if isinstance(output_path, str):
        output_path = Path(output_path)
    output_dir = output_path.parent if output_path else Path.cwd()
    if skip_mmd is False:
        events_csv_path = temp_dir / "logs" / "events.csv"
        if events_csv_path.exists():
            printer("Generating mermaid sequence diagram...")
            mmd_path = temp_dir / f"{flow_name}.mmd"
            generate_sequence_diagram(events_csv_path, mmd_path)
            if mmd_path.exists():
                shutil.copyfile(mmd_path, output_dir / f"{flow_name}.mmd")
    if output_path:
        destination_dir = output_path.parent
        destination_dir = (
            destination_dir
            / "waldiez_out"
            / datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        )
        destination_dir.mkdir(parents=True, exist_ok=True)
        # copy the contents of the temp dir to the destination dir
        printer(f"Copying the results to {destination_dir}")
        copy_results(
            temp_dir=temp_dir,
            output_path=output_path,
            output_dir=output_dir,
            destination_dir=destination_dir,
        )
    shutil.rmtree(temp_dir)


def copy_results(
    temp_dir: Path,
    output_path: Path,
    output_dir: Path,
    destination_dir: Path,
) -> None:
    """Copy the results to the output directory.

    Parameters
    ----------
    temp_dir : Path
        The temporary directory.
    output_path : Path
        The output path.
    output_dir : Path
        The output directory.
    destination_dir : Path
        The destination directory.
    """
    temp_dir.mkdir(parents=True, exist_ok=True)
    for item in temp_dir.iterdir():
        # skip cache files
        if (
            item.name.startswith("__pycache__")
            or item.name.endswith(".pyc")
            or item.name.endswith(".pyo")
            or item.name.endswith(".pyd")
            or item.name == ".cache"
        ):
            continue
        if item.is_file():
            # let's also copy the "tree of thoughts" image
            # to the output directory
            if item.name.endswith("tree_of_thoughts.png") or item.name.endswith(
                "reasoning_tree.json"
            ):
                shutil.copy(item, output_dir / item.name)
            shutil.copy(item, destination_dir)
        else:
            shutil.copytree(item, destination_dir / item.name)
    if output_path.is_file():
        if output_path.suffix == ".waldiez":
            output_path = output_path.with_suffix(".py")
        if output_path.suffix == ".py":
            src = temp_dir / output_path.name
            if src.exists():
                dst = destination_dir / output_path.name
                if dst.exists():
                    dst.unlink()
                shutil.copyfile(src, output_dir / output_path.name)


def get_printer() -> Callable[..., None]:
    """Get the printer function.

    Returns
    -------
    Callable[..., None]
        The printer function.
    """
    from autogen.io import IOStream  # type: ignore

    # noinspection PyUnboundLocalVariable
    printer = IOStream.get_default().print

    def safe_printer(*args: Any, **kwargs: Any) -> None:
        try:
            printer(*args, **kwargs)
        except UnicodeEncodeError:
            try:
                # Use the helper to get safely encoded message
                safe_msg, flush = get_what_to_print(*args, **kwargs)
                # Try printing the safe message
                # (without end since it's already included)
                printer(safe_msg, end="", flush=flush)
            except UnicodeEncodeError:
                # pylint: disable=too-many-try-statements
                try:
                    # If that still fails, write directly to buffer
                    msg, flush = get_what_to_print(*args, **kwargs)
                    safe_msg_bytes = msg.encode("utf-8", errors="replace")
                    sys.stdout.buffer.write(safe_msg_bytes)
                    if flush:
                        sys.stdout.buffer.flush()
                except Exception:  # pylint: disable=broad-exception-caught
                    error_msg = (
                        b"Could not print message due to encoding issues.\n"
                    )
                    sys.stderr.buffer.write(error_msg)
                    sys.stderr.buffer.flush()

    return safe_printer


def get_what_to_print(*args: object, **kwargs: object) -> tuple[str, bool]:
    """Get what to print.

    Parameters
    ----------
    args : object
        The print arguments.
    kwargs : object
        The keyword arguments.

    Returns
    -------
    tuple[str, bool]
        The message and whether to flush.
    """
    sep = kwargs.get("sep", " ")
    if not isinstance(sep, str):
        sep = " "
    end = kwargs.get("end", "\n")
    if not isinstance(end, str):
        end = "\n"
    flush = kwargs.get("flush", False)
    if not isinstance(flush, bool):
        flush = False
    msg = sep.join(str(arg) for arg in args) + end
    utf8_msg = msg.encode("utf-8", errors="replace").decode("utf-8")
    return utf8_msg, flush



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez agents model."""

from typing import Iterator

from pydantic import Field, model_validator
from typing_extensions import Annotated, Self

from ..common import WaldiezBase
from .agent import WaldiezAgent
from .assistant import WaldiezAssistant
from .captain import WaldiezCaptainAgent
from .group_manager import WaldiezGroupManager
from .rag_user_proxy import WaldiezRagUserProxy
from .reasoning import WaldiezReasoningAgent
from .user_proxy import WaldiezUserProxy


class WaldiezAgents(WaldiezBase):
    """Waldiez agents model.

    Attributes
    ----------
    userProxyAgents : list[WaldiezUserProxy]
        User proxy agents.
    assistantAgents : list[WaldiezAssistant]
        Assistant agents.
    ragUserProxyAgents : list[WaldiezRagUserProxy]
        RAG user proxy agents.
    reasoningAgents : list[WaldiezReasoningAgent]
        Reasoning agents.
    captainAgents : list[WaldiezCaptainAgent]
        Captain agents.
    groupManagerAgents : list[WaldiezGroupManager]
        Group manager agents.
    """

    userProxyAgents: Annotated[
        list[WaldiezUserProxy],
        Field(
            title="User Proxy Agents.",
            description="The User proxy agents in the flow.",
            default_factory=list,
        ),
    ] = []
    assistantAgents: Annotated[
        list[WaldiezAssistant],
        Field(
            title="Assistant Agents.",
            description="The assistant agents in the flow.",
            default_factory=list,
        ),
    ] = []
    ragUserProxyAgents: Annotated[
        list[WaldiezRagUserProxy],
        Field(
            title="RAG Users Proxy agents.",
            description="The RAG user proxy agents in the flow.",
            default_factory=list,
        ),
    ] = []
    reasoningAgents: Annotated[
        list[WaldiezReasoningAgent],
        Field(
            title="Reasoning Agents.",
            description="The Reasoning agents in the flow.",
            default_factory=list,
        ),
    ] = []
    captainAgents: Annotated[
        list[WaldiezCaptainAgent],
        Field(
            title="Captain Agents.",
            description="The Captain agents in the flow.",
            default_factory=list,
        ),
    ] = []
    groupManagerAgents: Annotated[
        list[WaldiezGroupManager],
        Field(
            title="Group Manager Agents.",
            description="The Group manager agents in the flow.",
            default_factory=list,
        ),
    ] = []

    @property
    def members(self) -> Iterator[WaldiezAgent]:
        """Get all agents.

        Yields
        ------
        WaldiezAgent
            The agents.
        """
        yield from self.userProxyAgents
        yield from self.assistantAgents
        yield from self.ragUserProxyAgents
        yield from self.reasoningAgents
        yield from self.captainAgents
        yield from self.groupManagerAgents

    @model_validator(mode="after")
    def validate_agents(self) -> Self:
        """Validate the agents.

        - At least two agents are required.
        - All the agent IDs must be unique.

        Returns
        -------
        WaldiezAgents
            The agents.

        Raises
        ------
        ValueError
            If the agents are invalid.
        """
        all_agent_ids = [agent.id for agent in self.members]
        if len(all_agent_ids) < 1:
            raise ValueError("At least one agent is required.")
        if len(all_agent_ids) != len(set(all_agent_ids)):
            raise ValueError("Agent IDs must be unique.")
        return self

    def validate_flow(self, model_ids: list[str], tool_ids: list[str]) -> None:
        """Validate the flow of the agents.

        - Validate the linked models (the referenced model ids must exist).
        - Validate the linked tools (the referenced tool ids must exist).
        - Validate the code execution (the referenced functions must exist).

        Parameters
        ----------
        model_ids : list[str]
            The list of model IDs.
        tool_ids : list[str]
            The list of tool IDs.

        Raises
        ------
        ValueError
            If the flow is invalid.
        """
        all_agent_ids = [agent.id for agent in self.members]
        for agent in self.members:
            agent.validate_linked_models(model_ids)
            agent.validate_linked_tools(tool_ids, agent_ids=all_agent_ids)
            agent.validate_code_execution(tool_ids=tool_ids)
            if agent.is_group_manager and isinstance(
                agent, WaldiezGroupManager
            ):
                agent.validate_initial_agent_id(all_agent_ids)
                agent.validate_transitions(agent_ids=all_agent_ids)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Base class to inherit from."""

from typing import Any

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class WaldiezBase(BaseModel):
    """Base model class to inherit from.

    It contains the default configuration for all models.
    It also `model_dumps` by alias by default.
    """

    model_config = ConfigDict(
        extra="ignore",
        # treat `toolId` as `tool_id`
        alias_generator=to_camel,
        # allow passing either `tool_id` or `toolId`
        populate_by_name=True,
        # allow setting any attribute after initialization
        frozen=False,
    )

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        """Dump the model to a dictionary.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

            The following are from the Pydantic `model_dump` method:
              - mode: str | Literal['json', 'python'] = 'json',
              - include: IncEx | None
              - exclude: IncEx | None
              - context: Any | None
              - by_alias: bool | None (None defaults to True)
              - exclude_unset: bool = False
              - exclude_defaults: bool = False
              - exclude_none: bool = False
              - round_trip: bool = False
              - warnings: bool | Literal['none', 'warn', 'error'] = True
              - fallback: ((Any) -> Any) | None = None
              - serialize_as_any: bool = False

        Returns
        -------
        dict[str, Any]
            The dictionary representation of the model.
        """
        by_alias = kwargs.pop("by_alias", None)
        mode = kwargs.pop("mode", None)
        if mode is None:  # pragma: no branch
            mode = "json"
        if by_alias is None:  # pragma: no branch
            by_alias = True
        if not isinstance(by_alias, bool):
            by_alias = True
        return super().model_dump(by_alias=by_alias, mode=mode, **kwargs)

    def model_dump_json(self, **kwargs: Any) -> str:
        """Dump the model to a JSON string.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

            The following are from the Pydantic `model_dump_json` method:
            - indent: int | None = None,
            - include: IncEx | None = None,
            - exclude: IncEx | None = None,
            - context: Any | None = None,
            - by_alias: bool | None = None, (None defaults to True)
            - exclude_unset: bool = False,
            - exclude_defaults: bool = False,
            - exclude_none: bool = False,
            - round_trip: bool = False,
            - warnings: bool | Literal['none', 'warn', 'error'] = True,
            - fallback: ((Any) -> Any) | None = None,
            - serialize_as_any: bool = False

        Returns
        -------
        str
            The JSON string.
        """
        by_alias = kwargs.pop("by_alias", None)
        if by_alias is None:
            by_alias = True
        if not isinstance(by_alias, bool):
            by_alias = True
        return super().model_dump_json(by_alias=by_alias, **kwargs)

    def __hash__(self) -> int:  # pragma: no cover
        """Return the hash of the object."""
        return id(self)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Core dataclasses for the exporting system."""

from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, Optional, TypeVar

from .enums import (
    ImportPosition,
)

# Generic type for extras
Extras = TypeVar("Extras")


# Sentinel for exporters that do not require extras
# pylint: disable=too-few-public-methods
class _NoExtrasType:
    """Sentinel for exporters that do not require extras."""

    def __repr__(self) -> str:
        """Get the string representation of NoExtras."""
        return "<NoExtras>"


NoExtras = _NoExtrasType()


# Core Data Structures
@dataclass
class ImportStatement:
    """Represents an import statement with its position."""

    statement: str
    position: ImportPosition = ImportPosition.THIRD_PARTY
    metadata: Optional[dict[str, Any]] = None

    def __hash__(self) -> int:
        """Hash based on the import statement.

        Returns
        -------
        int
            The hash value of the import statement.
        """
        return hash(self.statement)

    def __eq__(self, other: object) -> bool:
        """Check equality based on the import statement.

        Parameters
        ----------
        other : object
            The object to compare against.

        Returns
        -------
        bool
            True if the other object is an
            ImportStatement with the same statement.
        """
        if isinstance(other, ImportStatement):
            return self.statement == other.statement
        return False

    def __lt__(self, other: "ImportStatement") -> bool:
        """Enable sorting by position then statement.

        Parameters
        ----------
        other : ImportStatement
            The other import statement to compare against.

        Returns
        -------
        bool
            True if this import statement should come before the other.
        """
        if self.position.value != other.position.value:
            return self.position.value < other.position.value
        return self.statement < other.statement


@dataclass
class EnvironmentVariable:
    """Environment variable with metadata."""

    name: str
    value: str
    description: Optional[str] = None
    required: bool = True

    def __post_init__(self) -> None:
        """Validate environment variable."""
        if not self.name or not self.value:
            raise ValueError("Environment variable name and value are required")

    def __hash__(self) -> int:
        """Hash based on the environment variable name.

        Returns
        -------
        int
            The hash value of the environment variable name.
        """
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        """Check equality based on the environment variable name.

        Parameters
        ----------
        other : object
            The object to compare against.

        Returns
        -------
        bool
            True if the other object is an
            EnvironmentVariable with the same name.
        """
        if isinstance(other, EnvironmentVariable):
            return self.name == other.name
        return False

    def as_tuple(self) -> tuple[str, str]:
        """Get the environment variable as a tuple.

        Returns
        -------
        tuple[str, str]
            The environment variable as a tuple.
        """
        return (self.name, self.value)


# Instance Argument
@dataclass
class InstanceArgument:
    """Represents an instance argument for an agent, model or tool."""

    instance_id: str
    name: str
    value: Any
    tabs: int = 0
    tabs_length: int = 4  # Assuming 4 spaces per tab
    with_new_line_before: bool = False
    with_new_line_after: bool = False
    with_new_line_if_empty: bool = False
    skip_if_empty_string: bool = True
    comment: Optional[str] = None

    def has_content(self) -> bool:
        """Check if the instance argument has content.

        Returns
        -------
        bool
            True if the instance argument has content, otherwise False.
        """
        if self.skip_if_empty_string and isinstance(self.value, str):
            return bool(self.value.strip())
        return self.value is not None and self.value != ""

    def get_content(
        self,
        prepend_new_line: bool = False,
        append_new_line: bool = False,
    ) -> str:
        """Get the content representation of the instance argument.

        Parameters
        ----------
        prepend_new_line : bool, optional
            Whether to prepend a new line before the content,
            by default False.
        append_new_line : bool, optional
            Whether to append a new line at the end of the content,
            by default False.

        Returns
        -------
        str
            The formatted content string for the instance argument.
        """
        if (
            self.skip_if_empty_string
            and isinstance(self.value, str)
            and not self.value.strip()
        ):
            return "\n" if self.with_new_line_if_empty else ""
        space = " " * (self.tabs * self.tabs_length)
        content = f"{space}{self.name}={self.value}," + (
            f"  # {self.comment}" if self.comment else ""
        )
        if self.with_new_line_before or prepend_new_line:
            content = "\n" + content
        if self.with_new_line_after or append_new_line:
            content += "\n"
        return content

    def __hash__(self) -> int:
        """Hash based on the instance ID and name.

        Returns
        -------
        int
            The hash value of the instance argument.
        """
        return hash((self.instance_id, self.name))

    def __str__(self) -> str:
        """Get the string representation of the instance argument.

        Returns
        -------
        str
            The string representation of the instance argument.
        """
        return self.get_content()

    def __repr__(self) -> str:
        """Get the string representation of the instance argument.

        Returns
        -------
        str
            The string representation of the instance argument.
        """
        return self.get_content()


# Export Configuration
@dataclass
class ExportConfig:
    """Configuration for export operations.

    Attributes
    ----------
    output_extension : str
        The file extension for the exported content.
    is_async : bool
        Whether the exported content should be asynchronous.
    """

    name: str = "Waldiez Flow"
    description: str = (
        "Make AG2 Agents Collaborate: Drag, Drop, and Orchestrate with Waldiez"
    )
    requirements: list[str] = field(default_factory=list[str])
    tags: list[str] = field(default_factory=list[str])
    output_extension: str = "py"
    is_async: bool = False
    output_directory: Optional[str | Path] = None
    cache_seed: Optional[int] = None

    @property
    def for_notebook(self) -> bool:
        """Check if the export is intended for a notebook environment.

        Returns
        -------
        bool
            True if the output extension is 'ipynb', otherwise False.
        """
        return self.output_extension == "ipynb"

    def __post_init__(self) -> None:
        """Post-initialization validation."""
        if not self.name:
            raise ValueError("ExportConfig name cannot be empty")
        if not self.description:
            raise ValueError("ExportConfig description cannot be empty")
        if not self.output_extension:
            raise ValueError("ExportConfig output_extension cannot be empty")

    @classmethod
    def create(cls, **kwargs: Any) -> "ExportConfig":
        """Create a new ExportConfig instance with the provided values.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to initialize the ExportConfig.

        Returns
        -------
        ExportConfig
            A new instance of ExportConfig.
        """
        valid_fields = {f.name for f in fields(cls)}
        output_extension = kwargs.pop("output_extension", "py")
        for_notebook = kwargs.pop("for_notebook", output_extension == "ipynb")
        if for_notebook is True:
            output_extension = "ipynb"
        cache_seed = kwargs.pop("cache_seed", None)
        if cache_seed is not None and not isinstance(cache_seed, int):
            cache_seed = None
        return cls(
            cache_seed=cache_seed,
            output_extension=output_extension,
            **{k: v for k, v in kwargs.items() if k in valid_fields},
        )

    def update(self, **kwargs: Any) -> None:
        """Update the export configuration with new values.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments to update the configuration.

        Raises
        ------
        ValueError
            If an invalid configuration key is provided.
        """
        valid_fields = {f.name for f in fields(self)}
        for key, value in kwargs.items():
            if key in valid_fields:
                setattr(self, key, value)
        if (
            "for_notebook" in kwargs
            and isinstance(kwargs["for_notebook"], bool)
            and "output_extension" not in kwargs
        ):  # pragma: no cover
            self.output_extension = "ipynb" if kwargs["for_notebook"] else "py"



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Tool model."""

from pydantic import Field
from typing_extensions import Annotated

from ..common import WaldiezBase
from .tool_type import WaldiezToolType


class WaldiezToolData(WaldiezBase):
    """Waldiez Tool Data.

    Attributes
    ----------
    tool_type : WaldiezToolType
        The type of the tool: shared, custom, langchain, crewai.
    content : str
        The content (source code) of the tool.
    secrets : dict[str, str]
        The secrets (environment variables) of the tool.
    """

    tool_type: Annotated[
        WaldiezToolType,
        Field(
            "custom",
            alias="toolType",
            title="Tool Type",
            description=(
                "The type of the tool: shared, custom, langchain, crewai."
            ),
        ),
    ] = "custom"
    content: Annotated[
        str,
        Field(
            ...,
            title="Content",
            description="The content (source code) of the tool.",
        ),
    ]
    secrets: Annotated[
        dict[str, str],
        Field(
            default_factory=dict,
            title="Secrets",
            description="The secrets (environment variables) of the tool.",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez model extra requirements."""

from typing import Iterator, Set

from .model import WaldiezModel


def get_models_extra_requirements(
    models: Iterator[WaldiezModel], autogen_version: str
) -> Set[str]:
    """Get the models extra requirements.

    Parameters
    ----------
    models : list[WaldiezModel]
        The models.
    autogen_version : str
        The autogen version.

    Returns
    -------
    list[str]
        The models extra requirements.
    """
    model_requirements: Set[str] = set()
    # ref: https://github.com/ag2ai/ag2/blob/main/pyproject.toml
    models_with_additional_requirements = [
        "together",
        "gemini",
        "mistral",
        "groq",
        "anthropic",
        "cohere",
        "bedrock",
    ]
    for model in models:
        for requirement in model.requirements:
            model_requirements.add(requirement)
        if model.data.api_type == "google":
            model_requirements.add(f"ag2[gemini]=={autogen_version}")
            continue
        if model.data.api_type in models_with_additional_requirements:
            model_requirements.add(
                f"ag2[{model.data.api_type}]=={autogen_version}"
            )
    return model_requirements



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods,too-many-instance-attributes
# flake8: noqa: E501
"""Flow specific extras module."""

from dataclasses import dataclass, field
from typing import Optional

# pylint: disable=import-error,line-too-long
# pyright: reportMissingImports=false
try:
    from waldiez._version import __version__ as waldiez_version  # type: ignore[unused-ignore, import-not-found, import-untyped]  # noqa
except ImportError:  # pragma: no cover
    import warnings

    # .gitingored (generated by hatch)
    warnings.warn(
        "Importing __version__ failed. Using 'dev' as version.",
        stacklevel=2,
    )
    waldiez_version = "dev"

from ..result import ExportResult
from ..types import ExportConfig
from .base import BaseExtras


@dataclass
class FlowExtras(BaseExtras):
    """Extras for flow exporter."""

    # Flow metadata
    flow_name: str = ""
    description: str = ""
    config: ExportConfig = field(default_factory=ExportConfig)
    version: str = waldiez_version  # pyright: ignore

    # Sub-exporter results
    tools_result: Optional[ExportResult] = None
    models_result: Optional[ExportResult] = None
    agents_result: Optional[ExportResult] = None
    chats_result: Optional[ExportResult] = None

    # Generated script parts
    header_content: str = ""
    main_function: str = ""
    after_run: str = ""
    execution_code: str = ""

    def has_specific_content(self) -> bool:
        """Check if the flow extras contain specific content.

        Returns
        -------
        bool
            True if any specific content is present, False otherwise.
        """
        return bool(
            self.flow_name
            or self.description
            or self.version
            or self.tools_result
            or self.models_result
            or self.agents_result
            or self.chats_result
            or self.header_content
            or self.main_function
            or self.execution_code
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""User input data models and validation."""

import json
import os
from pathlib import Path
from typing import Any, Union
from urllib.parse import urlparse

from pydantic import BaseModel, Field, field_validator
from typing_extensions import Annotated

from ..utils import detect_media_type
from .constants import CONTENT_MAPPING, ContentMappingEntry, MediaContent
from .content.audio import AudioMediaContent
from .content.base import (
    AudioContent,
    FileContent,
    ImageContent,
    VideoContent,
)
from .content.file import FileMediaContent
from .content.image import ImageMediaContent, ImageUrlMediaContent
from .content.text import TextMediaContent
from .content.video import VideoMediaContent


class UserInputData(BaseModel):
    """User's input data model."""

    content: Annotated[
        Union[MediaContent, list[MediaContent]],
        Field(
            description="The content of the input data.",
            title="Content",
        ),
    ]

    def __str__(self) -> str:
        """Get the string representation of the content."""
        return self.to_string()

    def __repr__(self) -> str:
        """Get the string representation of the UserInputData."""
        return f"UserInputData(content={self.to_string()})"

    def to_string(
        self,
        uploads_root: Path | None = None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        if isinstance(self.content, list):
            return " ".join(
                [
                    item.to_string(uploads_root, base_name)  # pyright: ignore
                    for item in self.content
                ]
            )
        return self.content.to_string(uploads_root, base_name)

    @classmethod
    def content_from_string(cls, value: str) -> MediaContent:
        """Convert a string to the appropriate MediaContent type.

        Parameters
        ----------
        value : str
            The input string

        Returns
        -------
        MediaContent
            the appropriate MediaContent.
        """
        try:
            parsed = json.loads(value)
        except json.JSONDecodeError:
            return TextMediaContent(type="text", text=value)
        if isinstance(parsed, str):
            try:
                parsed = json.loads(parsed)
            except json.JSONDecodeError:
                return TextMediaContent(type="text", text=parsed)
            if isinstance(parsed, str):
                return TextMediaContent(type="text", text=parsed)
        return cls.validate_content(parsed)

    @classmethod
    def content_from_dict(cls, value: dict[str, Any]) -> MediaContent:
        """Convert a dictionary to the appropriate MediaContent type.

        Parameters
        ----------
        value : dict[str, Any]
            The input dictionary

        Returns
        -------
        MediaContent
            the appropriate MediaContent.

        Raises
        ------
        ValueError
            If the content type is not supported or
            if any required field is missing.
        """
        content_type = detect_media_type(value)

        # Get the mapping for the detected content type
        if content_type not in CONTENT_MAPPING:
            raise ValueError(f"Unsupported content type: {content_type}")

        # Get the mapping for the content type
        mapping = CONTENT_MAPPING[content_type]
        return cls._build_media_content(value, mapping, content_type)

    @classmethod
    def _build_media_content(
        cls,
        value: dict[str, Any],
        mapping: ContentMappingEntry,
        content_type: str,
    ) -> MediaContent:
        """Try to construct the appropriate MediaContent from the mapping."""
        for field in mapping["fields"]:
            if field not in value:
                continue

            raw_val = value[field]

            if field == mapping["required_field"]:
                # if we have direct mapping to the required field,
                # let's try to instantiate the class directly
                try:
                    return mapping["cls"](**{field: raw_val})
                except ValueError:
                    pass  # let's try to convert it

            converted = cls._convert_simple_content(
                raw_val, mapping["required_field"]
            )
            if converted is not None:  # pragma: no branch
                return mapping["cls"](
                    **{mapping["required_field"]: converted}  # type: ignore
                )

        raise ValueError(
            "Missing required field for content type "
            f"'{content_type}' in: {value}"
        )

    @staticmethod
    def _convert_simple_content(
        raw_val: Any, target_field: str
    ) -> ImageContent | VideoContent | AudioContent | FileContent | None:
        """Convert a simple string to the appropriate content, if applicable."""
        if not isinstance(raw_val, str):  # pragma: no cover
            return None

        if target_field in ("image_url", "image"):
            return ImageContent(url=raw_val)
        if target_field == "video":
            return VideoContent(url=raw_val)
        if target_field == "audio":
            return AudioContent(url=raw_val)
        if target_field == "file":
            filename = extract_filename_from_path(raw_val)
            return FileContent(name=filename, url=raw_val)

        return None  # pragma: no cover

    @field_validator("content", mode="before")
    @classmethod
    def validate_content(cls, v: Any) -> MediaContent:  # noqa: C901,D102
        """Validate the input data content.

        Parameters
        ----------
        v: Any
            The input data content

        Returns
        -------
        MediaContent
            The validated content

        Raises
        ------
        ValueError
            If the content is not valid.
        """
        if isinstance(
            v,
            (
                TextMediaContent,
                ImageMediaContent,
                ImageUrlMediaContent,
                VideoMediaContent,
                AudioMediaContent,
                FileMediaContent,
            ),
        ):
            return v

        # If it's a string, check if it is a dumped one
        if isinstance(v, str):
            return cls.content_from_string(v)

        # If it's a dictionary, check if it has a 'type' field
        if isinstance(v, dict):
            return cls.content_from_dict(v)  # pyright: ignore

        # If it's a list
        if isinstance(v, list):
            return [cls.validate_content(item) for item in v]  # type: ignore

        # Default fallback
        return TextMediaContent(type="text", text=str(v))


def extract_filename_from_path(path_or_url: str) -> str:
    """Extract the filename from a given path or URL.

    Parameters
    ----------
    path_or_url: str
        The path or URL from which to extract the filename.

    Returns
    -------
    str
        The extracted filename.
    """
    if "://" in path_or_url:
        # It's a URL - parse it properly
        parsed = urlparse(path_or_url)
        # Extract filename from the path component
        filename = os.path.basename(parsed.path)

        # Handle edge cases where path might be empty or end with /
        if not filename and parsed.path.endswith("/"):
            # Try to get the last directory name
            path_parts = [p for p in parsed.path.split("/") if p]
            filename = path_parts[-1] if path_parts else parsed.netloc
        elif not filename:
            # Fallback to using netloc (domain) if no path
            filename = parsed.netloc

        return filename
    # else:
    # Local file path
    return os.path.basename(path_or_url)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common utilities for exporting chats."""

from typing import Optional

from waldiez.models import WaldiezAgent, WaldiezChat


def get_chat_message_string(
    sender: WaldiezAgent,
    chat: WaldiezChat,
    chat_names: dict[str, str],
) -> tuple[str, Optional[str]]:
    """Get the agent's message as a string.

    Parameters
    ----------
    sender : WaldiezAgent
        The sender.
    chat : WaldiezChat
        The chat.
    chat_names : dict[str, str]
        A mapping of chat id to chat name with all the chats in the flow.

    Returns
    -------
    tuple[str, Optional[str]]
        If the message is a string, the message content and None.
        If the message is a method, the method name and the method content.
        If the message is None, 'None' and None.
    """
    if not chat.message or chat.message.type == "none":
        return "None", None
    if chat.message.type == "string":
        if chat.message.content is None:  # pragma: no cover
            # should be coverred previousliy on pydantic validation
            return "None", None
        return chat.message.content, None

    is_rag_with_carryover = sender.is_rag_user and chat.message.use_carryover
    chat_name = chat_names[chat.id]
    function_content, function_name = chat.get_message_function(
        name_suffix=chat_name,
        is_rag=is_rag_with_carryover,
    )
    return function_name, function_content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Chats exporter."""

from .exporter import ChatsExporter
from .factory import create_chats_exporter

__all__ = [
    "ChatsExporter",
    "create_chats_exporter",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Assistant agent model."""

from .assistant import WaldiezAssistant
from .assistant_data import WaldiezAssistantData

__all__ = ["WaldiezAssistant", "WaldiezAssistantData"]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Structured I/O stream for JSON-based communication over stdin/stdout."""

import json
import queue
import sys
import threading
from getpass import getpass
from pathlib import Path
from typing import Any
from uuid import uuid4

from autogen.events import BaseEvent  # type: ignore
from autogen.io import IOStream  # type: ignore

from .models import (
    PrintMessage,
    UserInputData,
    UserInputRequest,
    UserResponse,
)
from .utils import (
    gen_id,
    get_image,
    is_json_dumped,
    now,
    try_parse_maybe_serialized,
)


class StructuredIOStream(IOStream):
    """Structured I/O stream using stdin and stdout."""

    uploads_root: Path | None = None

    def __init__(
        self, timeout: float = 120, uploads_root: Path | str | None = None
    ) -> None:
        self.timeout = timeout
        if uploads_root is not None:
            self.uploads_root = Path(uploads_root).resolve()
            if not self.uploads_root.exists():
                self.uploads_root.mkdir(parents=True, exist_ok=True)

    # noinspection PyMethodMayBeStatic
    # pylint: disable=no-self-use
    def print(self, *args: Any, **kwargs: Any) -> None:
        """Structured print to stdout.

        Parameters
        ----------
        args : Any
            The data to print.
        kwargs : Any
        """
        sep = kwargs.get("sep", " ")
        end = kwargs.get("end", "\n")
        message = sep.join(map(str, args))
        is_dumped = is_json_dumped(message)
        if is_dumped and end.endswith("\n"):
            # If the message is already JSON-dumped,
            # let's try not to double dump it
            message = json.loads(message)
            payload: dict[str, Any] = {
                "type": "print",
                "id": uuid4().hex,
                "timestamp": now(),
                "data": message,
            }
        else:
            message += end
            payload = PrintMessage(
                id=uuid4().hex,
                timestamp=now(),
                data=message,
            ).model_dump(mode="json")
        flush = kwargs.get("flush", True)
        payload_type = kwargs.get("type", "print")
        payload["type"] = payload_type
        print(json.dumps(payload), flush=flush)

    def input(self, prompt: str = "", *, password: bool = False) -> str:
        """Structured input from stdin.

        Parameters
        ----------
        prompt : str, optional
            The prompt to display. Defaults to "".
        password : bool, optional
            Whether to read a password. Defaults to False.

        Returns
        -------
        str
            The line read from the input stream.
        """
        request_id = uuid4().hex
        prompt = prompt or "> "

        self._send_input_request(prompt, request_id, password)
        user_input_raw = self._read_user_input(prompt, password, request_id)
        response = self._handle_user_input(user_input_raw, request_id)
        user_response = response.to_string(
            uploads_root=self.uploads_root,
            base_name=request_id,
        )
        return user_response

    # noinspection PyMethodMayBeStatic
    # pylint: disable=no-self-use
    def send(self, message: BaseEvent) -> None:
        """Structured sending of a BaseEvent.

        Parameters
        ----------
        message : BaseEvent
            The message to send.
        """
        message_dump = message.model_dump(mode="json")
        if message_dump.get("type") == "text":
            content_block = message_dump.get("content")
            if (
                isinstance(content_block, dict)
                and "content" in content_block
                and isinstance(content_block["content"], str)
            ):
                inner_content = content_block["content"]
                content_block["content"] = try_parse_maybe_serialized(
                    inner_content
                )
        print(json.dumps(message_dump), flush=True, file=sys.stdout)

    # noinspection PyMethodMayBeStatic
    # pylint: disable=no-self-use
    def _send_input_request(
        self,
        prompt: str,
        request_id: str,
        password: bool,
    ) -> None:
        payload = UserInputRequest(
            request_id=request_id,
            prompt=prompt,
            password=password,
        ).model_dump(mode="json")
        print(json.dumps(payload), flush=True)

    def _read_user_input(
        self,
        prompt: str,
        password: bool,
        request_id: str,
    ) -> str:
        input_queue: queue.Queue[str] = queue.Queue()

        def read_input() -> None:
            """Read user input from stdin."""
            try:
                user_input = (
                    getpass(prompt).strip()
                    if password
                    else input(prompt).strip()
                )
                input_queue.put(user_input)
            except EOFError:
                input_queue.put("")

        input_thread = threading.Thread(target=read_input, daemon=True)
        input_thread.start()

        try:
            return input_queue.get(timeout=self.timeout)
        except queue.Empty:
            self._send_timeout_message(request_id)
            return ""

    def _send_timeout_message(self, request_id: str) -> None:
        timeout_payload = {
            "id": gen_id(),
            "type": "timeout",
            "request_id": request_id,
            "timestamp": now(),
            "data": f"No input received after {self.timeout} seconds.",
        }
        print(json.dumps(timeout_payload), flush=True, file=sys.stderr)

    def _handle_user_input(
        self, user_input_raw: str, request_id: str
    ) -> UserResponse:
        """Handle user input and return the appropriate response.

        Parameters
        ----------
        user_input_raw : str
            The raw user input string.
        request_id : str
            The request ID to match against.

        Returns
        -------
        UserResponse
            The structured user response.
        """
        user_input = self._load_user_input(user_input_raw)
        if isinstance(user_input, str):
            return UserResponse(data=user_input, request_id=request_id)
        return self._parse_user_input(user_input, request_id)

    @staticmethod
    def _load_user_input(user_input_raw: str) -> str | dict[str, Any]:
        """Load user input from a raw string.

        Parameters
        ----------
        user_input_raw : str
            The raw user input string.

        Returns
        -------
        str | dict[str, Any]
            The loaded user input, either as a string or a dictionary.
        """
        response: str | dict[str, Any] = user_input_raw
        try:
            # Attempt to parse the input as JSON
            response = json.loads(user_input_raw)
        except json.JSONDecodeError:
            # If it's not valid JSON, return as is
            # This allows for backwards compatibility with raw text input
            return user_input_raw
        if isinstance(response, str):
            # double inner dumped?
            try:
                response = json.loads(response)
            except json.JSONDecodeError:
                # If it's not valid JSON, return as is
                return response
        if not isinstance(response, dict):
            return str(response)
        if "data" in response and isinstance(response["data"], str):
            # double inner dumped?
            try:
                response["data"] = json.loads(response["data"])
            except json.JSONDecodeError:
                pass
        return response

    def _parse_user_input(
        self, user_input: dict[str, Any], request_id: str
    ) -> UserResponse:
        """Parse user input and return the appropriate response.

        Parameters
        ----------
        user_input : str
            The raw user input string.
        request_id : str
            The request ID to match against.

        Returns
        -------
        UserResponse
            The structured user response.
        """
        # Load the user input
        if user_input.get("request_id") == request_id:
            # We have a valid response to our request
            data = user_input.get("data")
            if not data:
                # let's check if text|image keys are sent (outside data)
                if "image" in user_input or "text" in user_input:
                    return UserResponse(
                        request_id=request_id,
                        data=self._format_multimedia_response(
                            request_id=request_id, data=user_input
                        ),
                    )
            if isinstance(data, list):
                return self._handle_list_response(
                    data,  # pyright: ignore
                    request_id=request_id,
                )
            if not data or not isinstance(data, (str, dict)):
                # No / invalid data provided in the response
                return UserResponse(
                    request_id=request_id,
                    data="",
                )
            # Process different data types
            if isinstance(data, str):
                # double inner dumped?
                data = self._load_user_input(data)
            if isinstance(data, dict):
                return UserResponse(
                    data=self._format_multimedia_response(
                        request_id=request_id,
                        data=data,  # pyright: ignore
                    ),
                    request_id=request_id,
                )
            # For other types (numbers, bools ,...),
            #  let's just convert to string
            return UserResponse(
                data=str(data), request_id=request_id
            )  # pragma: no cover
        # This response doesn't match our request_id, log and return empty
        self._log_mismatched_response(request_id, user_input)
        return UserResponse(
            request_id=request_id,
            data="",
        )

    def _handle_list_response(
        self,
        data: list[dict[str, Any]],
        request_id: str,
    ) -> UserResponse:
        if len(data) == 0:  # pyright: ignore
            # Empty list, return empty response
            return UserResponse(
                request_id=request_id,
                data="",
            )

        input_data: list[UserInputData] = []
        for entry in data:  # pyright: ignore
            # pylint: disable=broad-exception-caught
            try:
                content = UserInputData.model_validate(entry)
                input_data.append(content)
            except Exception as error:  # pragma: no cover
                print({"type": "error", "message": str(error)}, file=sys.stderr)
                continue
        if not input_data:  # pragma: no cover
            # No valid data in the list, return empty response
            return UserResponse(
                request_id=request_id,
                data="",
            )
        return UserResponse(
            request_id=request_id,
            data=input_data,
        )

    @staticmethod
    def _log_mismatched_response(expected_id: str, response: Any) -> None:
        """Log information about mismatched response IDs.

        Parameters
        ----------
        expected_id : str
            The request ID we were expecting
        response : Any
            The response received
        """
        # Create a log message
        log_payload: dict[str, Any] = {
            "type": "warning",
            "id": uuid4().hex,
            "timestamp": now(),
            "data": {
                "message": (
                    "Received response with mismatched request_id. "
                    f"Expected: {expected_id}"
                ),
                "details": {
                    "expected_id": expected_id,
                    "received": str(response)[:100]
                    + ("..." if len(str(response)) > 100 else ""),
                },
            },
        }
        # Print to stderr to avoid interfering with stdout communication
        print(json.dumps(log_payload), file=sys.stderr)

    def _format_multimedia_response(
        self,
        data: dict[str, Any],
        request_id: str | None = None,
    ) -> str:
        """Format a multimedia response dict into a string with image tags.

        Parameters
        ----------
        data : dict[str, Any]
            The data dictionary containing "image" and "text" keys.
        request_id : str | None
            The input request ID, if available.

        Returns
        -------
        str
            The formatted string with image tags and text.
        """
        result: list[str] = []
        if "content" in data and isinstance(data["content"], dict):
            return self._format_multimedia_response(
                data=data["content"],  # pyright: ignore
                request_id=request_id,
            )

        # Handle image if present
        if "image" in data and data["image"]:
            image_data = data["image"]
            image = get_image(self.uploads_root, image_data, request_id)
            img_tag = f"<img {image}>"
            result.append(img_tag)

        # Handle text if present
        if "text" in data and data["text"]:
            result.append(str(data["text"]))

        # If neither image nor text, return empty string
        if not result:
            return ""
        # Join with a space
        return " ".join(result)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Factory function for creating a ToolsExporter instance."""

from pathlib import Path
from typing import Optional, Union

from waldiez.models import WaldiezAgent, WaldiezTool

from ..core import ExporterContext, get_default_exporter_context
from .exporter import ToolsExporter


def create_tools_exporter(
    flow_name: str,
    agents: list[WaldiezAgent],
    agent_names: dict[str, str],
    tools: list[WaldiezTool],
    tool_names: dict[str, str],
    output_dir: Optional[Union[str, Path]] = None,
    context: Optional[ExporterContext] = None,
) -> ToolsExporter:
    """Create a tools exporter.

    Parameters
    ----------
    flow_name : str
        The name of the flow.
    agents : list[WaldiezAgent]
        The agents that use tools.
    agent_names : dict[str, str]
        Mapping of agent IDs to names.
    tools : list[WaldiezTool]
        The tools to export.
    tool_names : dict[str, str]
        Mapping of tool IDs to names.
    output_dir : Optional[Union[str, Path]], optional
        Output directory for generated files, by default None
    context : Optional[ExporterContext], optional
        Exporter context with dependencies, by default None

    Returns
    -------
    ToolsExporter
        The created tools exporter.
    """
    if context is None:
        context = get_default_exporter_context()
    return ToolsExporter(
        flow_name=flow_name,
        agents=agents,
        agent_names=agent_names,
        tools=tools,
        tool_names=tool_names,
        output_dir=output_dir,
        context=context,
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez entrypoint when called as a module."""

from .cli import app

if __name__ == "__main__":
    app()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=import-outside-toplevel,reimported
"""Environment related utilities."""

import os
import sys
from typing import Generator


def in_virtualenv() -> bool:
    """Check if we are inside a virtualenv.

    Returns
    -------
    bool
        True if inside a virtualenv, False otherwise.
    """
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix")
        and os.path.realpath(sys.base_prefix) != os.path.realpath(sys.prefix)
    )


def is_root() -> bool:  # pragma: no cover  # os specific
    """Check if the script is running as root/administrator.

    Returns
    -------
    bool
        True if running as root/administrator, False otherwise.
    """
    # pylint: disable=import-outside-toplevel,line-too-long,no-member
    if os.name == "nt":
        # noinspection PyBroadException
        # pylint: disable=broad-exception-caught
        try:
            import ctypes

            return ctypes.windll.shell32.IsUserAnAdmin() != 0  # type: ignore[unused-ignore,attr-defined]  # noqa: E501
        except Exception:
            return False
    else:
        return os.getuid() == 0


def refresh_environment() -> None:
    """Refresh the environment."""
    # a group chat without a user agent
    # creates a new user (this has a default code execution with docker)
    # captain also generates new agents that also have
    # default code execution with docker
    # temp (until we handle/detect docker setup)
    os.environ["AUTOGEN_USE_DOCKER"] = "0"
    try_handle_the_np_thing()


def try_handle_the_np_thing() -> None:
    """Try to handle the numpy deprecation warning."""
    # we might get:
    # module 'numpy' has no attribute '_no_nep50_warning'
    # (sentnence_transformers?)
    # in autogen/agentchat/contrib/captainagent/tool_retriever.py
    os.environ["NEP50_DEPRECATION_WARNING"] = "0"
    os.environ["NEP50_DISABLE_WARNING"] = "1"
    os.environ["NPY_PROMOTION_STATE"] = "weak"
    import numpy as np

    if not hasattr(np, "_no_pep50_warning"):
        import contextlib

        @contextlib.contextmanager
        def _np_no_nep50_warning() -> Generator[None, None, None]:
            """Avoid no_nep50 warning.

            Yields
            ------
            None
                Nothing.
            """
            yield  # pragma: no cover

        setattr(np, "_no_pep50_warning", _np_no_nep50_warning)  # noqa


def set_env_vars(flow_env_vars: list[tuple[str, str]]) -> dict[str, str]:
    """Set environment variables and return the old ones (if any).

    Parameters
    ----------
    flow_env_vars : list[tuple[str, str]]
        The environment variables to set.

    Returns
    -------
    dict[str, str]
        The old environment variables.
    """
    old_vars: dict[str, str] = {}
    for var_key, var_value in flow_env_vars:
        if var_key:
            current = os.environ.get(var_key, "")
            old_vars[var_key] = current
            os.environ[var_key] = var_value
    return old_vars


def reset_env_vars(old_vars: dict[str, str]) -> None:
    """Reset the environment variables.

    Parameters
    ----------
    old_vars : dict[str, str]
        The old environment variables.
    """
    for var_key, var_value in old_vars.items():
        if not var_value:
            os.environ.pop(var_key, "")
        else:
            os.environ[var_key] = var_value



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Reasoning agent model."""

from typing import Any

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgent
from .reasoning_agent_data import WaldiezReasoningAgentData


class WaldiezReasoningAgent(WaldiezAgent):
    """Reasoning agent model."""

    agent_type: Annotated[  # pyright: ignore
        Literal["reasoning"],
        Field(
            "reasoning",
            title="Agent type",
            description="The agent type in a graph: 'reasoning'",
            alias="agentType",
        ),
    ] = "reasoning"
    data: Annotated[  # pyright: ignore
        WaldiezReasoningAgentData,
        Field(
            title="Data",
            description="The reasoning agent's data",
            default_factory=WaldiezReasoningAgentData,  # pyright: ignore
        ),
    ]

    def get_reasoning_config(self) -> dict[str, Any]:
        """Get the reasoning configuration based on its method.

        Returns
        -------
        dict
            The reasoning configuration.
        """
        return self.data.get_reasoning_config()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Utility functions for the waldiez.io package."""

import ast
import json
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Literal, Union

from autogen.agentchat.contrib.img_utils import get_pil_image  # type: ignore

MessageType = Literal[
    "input_request",
    "input_response",
    "print",
    "input",
]
"""Possible message types for the structured I/O stream."""

MediaType = Union[
    Literal["text"],
    Literal["image"],
    Literal["image_url"],
    Literal["video"],
    Literal["audio"],
    Literal["file"],
]
"""Possible media types for the structured I/O stream."""


def gen_id() -> str:
    """Generate a unique identifier.

    Returns
    -------
    str
        A unique id
    """
    return str(uuid.uuid4())


def now() -> str:
    """Get the current time as an ISO string.

    Returns
    -------
    str
        The current time as an ISO string.
    """
    return datetime.now().isoformat()


def detect_media_type(value: dict[str, Any]) -> MediaType:
    """Detect mediatype from dict.

    Either using the 'type' field or by checking the keys.

    Parameters
    ----------
    value : dict[str, Any]
        The input dictionary

    Returns
    -------
    MediaType
        The detected media type

    Raises
    ------
    ValueError
        If the media type is not valid or not found.
    """
    valid_types = (
        "text",
        "image",
        "image_url",
        "video",
        "audio",
        "file",
    )
    if "type" in value:
        content_type = value["type"]
        if content_type not in valid_types:
            raise ValueError(f"Invalid media type: {content_type}")
        return content_type
    for valid_type in valid_types:
        if valid_type in value:
            return valid_type  # type: ignore
    raise ValueError(f"No type in value: {value}.")


def get_image(
    uploads_root: Path | None,
    image_data: str,
    base_name: str | None = None,
) -> str:
    """Store the image data in a file and return the file path.

    Parameters
    ----------
    uploads_root : Path | None
        The root directory for storing images, optional.
    image_data : str
        The base64-encoded image data.
    base_name : str | None
        The base name for the image file, optional.

    Returns
    -------
    str
        The file path of the stored image.
    """
    if uploads_root:
        # noinspection PyBroadException
        # pylint: disable=broad-exception-caught
        try:
            pil_image = get_pil_image(image_data)
        except BaseException:
            return image_data
        if not base_name:
            base_name = uuid.uuid4().hex
        file_name = f"{base_name}.png"
        file_path = uploads_root / file_name
        pil_image.save(file_path, format="PNG")
        return str(file_path)
    return image_data


def is_json_dumped(value: str) -> bool:
    """Check if a string is JSON-dumped.

    Parameters
    ----------
    value : str
        The string to check.

    Returns
    -------
    bool
        True if the string is JSON-dumped, False otherwise.
    """
    try:
        parsed = json.loads(value)
        # If we can parse it as JSON and it's not a string,
        # we consider it JSON-dumped
        return not isinstance(parsed, str)
    except json.JSONDecodeError:
        return False


def try_parse_maybe_serialized(value: str) -> Any:
    """Parse a string that may be JSON or Python serialized.

    Returns the parsed object if successful, or the original string otherwise.

    Parameters
    ----------
    value : str
        The string to parse.

    Returns
    -------
    Any
        The parsed object or the original string if parsing fails.
    """
    for parser in (json.loads, ast.literal_eval):
        # pylint: disable=broad-exception-caught, too-many-try-statements
        try:
            parsed: dict[str, Any] | list[Any] | str = parser(value)
            # Normalize: if it's a single-item list of a string
            # return the string
            if (
                isinstance(parsed, list)
                and len(parsed) == 1
                and isinstance(parsed[0], str)
            ):
                return parsed[0]
            return parsed
        except Exception:
            pass  # Try next parser
    return value  # Return original if all parsing fails



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utility functions for exporting sequential chats."""

from typing import Callable

from waldiez.models import (
    WaldiezAgentConnection,
    WaldiezChatMessage,
    WaldiezRagUserProxy,
)

from .common import get_chat_message_string


def export_sequential_chat(
    main_chats: list[WaldiezAgentConnection],
    chat_names: dict[str, str],
    agent_names: dict[str, str],
    serializer: Callable[..., str],
    tabs: int,
    is_async: bool,
    skip_cache: bool,
) -> tuple[str, str]:
    r"""Get the chats content, when there are more than one chats in the flow.

    Parameters
    ----------
    main_chats : list[WaldiezAgentConnection]
        The main chats.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    agent_names : dict[str, str]
        A mapping of agent id to agent name.
    serializer : Callable[..., str]
        The serializer function to escape quotes in a string.
    tabs : int
        The number of tabs to use for indentation.
    is_async : bool
        Whether the chat is asynchronous.
    skip_cache : bool
        Whether to skip the cache argument.

    Returns
    -------
    tuple[str, str]
        The main chats content and additional methods string if any.

    Example
    -------
    ```python
    >>> from waldiez.models import (
    ...     WaldiezAgent,
    ...     WaldiezChat,
    ...     WaldiezChatData,
    ...     WaldiezChatMessage,
    ... )
    >>> chat1 = WaldiezChat(
    ...     id="wc-1",
    ...     name="chat1",
    ...     description="A chat between two agents.",
    ...     tags=["chat", "chat1"],
    ...     requirements=[],
    ...     data=WaldiezChatData(
    ...         sender="wa-1",
    ...         recipient="wa-2",
    ...         position=0,
    ...         message=WaldiezChatMessage(
    ...             type="string",
    ...             content="Hello, how are you?",
    ...         ),
    ...     ),
    ... )
    >>> chat2 = WaldiezChat(
    ...     id="wc-2",
    ...     name="chat2",
    ...     description="A chat between two agents.",
    ...     tags=["chat", "chat2"],
    ...     requirements=[],
    ...     data=WaldiezChatData(
    ...         sender="wa-2",
    ...         recipient="wa-1",
    ...         position=1,
    ...         message=WaldiezChatMessage(
    ...             type="string",
    ...             content="I am good, thank you. How about you?",
    ...         ),
    ...     ),
    ... )
    >>> agent_names = {"wa-1": "agent1", "wa-2": "agent2"}
    >>> chat_names = {"wc-1": "chat1", "wc-2": "chat2"}
    >>> serializer = lambda x: x.replace('"', "\"").replace("\n", "\\n")
    >>>  export_sequential_chat(
    ...     main_chats=[(chat1, agent1, agent2), (chat2, agent2, agent1)],
    ...     chat_names=chat_names,
    ...     agent_names=agent_names,
    ...     serializer=serializer,
    ...     tabs=0,
    ...     is_async=False,
    ... )
    results = initiate_chats([
        {
            "sender": agent1,
            "recipient": agent2,
            "message": "Hello, how are you?",
        },
        {
            "sender": agent2,
            "recipient": agent1,
            "message": "I am good, thank you. How about you?",
        },
    ])
    ```
    """
    tab = "    " * tabs if tabs > 0 else ""
    content = "\n"
    additional_methods_string = ""
    content += _get_initiate_chats_line(tab, is_async)
    for connection in main_chats:
        chat_string, additional_methods = _get_chat_dict_string(
            chat_names=chat_names,
            connection=connection,
            agent_names=agent_names,
            serializer=serializer,
            tabs=tabs + 1,
            skip_cache=skip_cache,
        )
        additional_methods_string += additional_methods
        content += "\n" + f"{tab}    {chat_string}"
    content += "\n" + "    " * tabs + "])\n"
    return content, additional_methods_string


def _get_chat_dict_string(
    connection: WaldiezAgentConnection,
    chat_names: dict[str, str],
    agent_names: dict[str, str],
    serializer: Callable[..., str],
    tabs: int,
    skip_cache: bool,
) -> tuple[str, str]:
    """Get a chat dictionary string.

    If the chat message is a separate method and not a string or a lambda,
    we return the method string (definition and body) as well as the rest
    of the arguments.

    Parameters
    ----------
    connection : WaldiezAgentConnection
        The connection object containing the chat and agents.
    chat_names : dict[str, str]
        A mapping of chat id to chat name.
    agent_names : dict[str, str]
        A mapping of agent id to agent name.
    serializer : Callable[[str], str]
        The function to serialize the dictionaries or lists.
    tabs : int
        The number of tabs to use for indentation.
    skip_cache : bool
        Whether to skip the cache argument.

    Returns
    -------
    tuple[str, str]
        The chat dictionary string and additional methods string if any.
    """
    tab = "    " * tabs
    chat = connection["chat"]
    sender = connection["source"]
    chat_string = _get_chat_strig_start(
        connection=connection,
        agent_names=agent_names,
        serializer=serializer,
        tabs=tabs,
        skip_cache=skip_cache,
    )
    if (
        sender.is_rag_user
        and isinstance(sender, WaldiezRagUserProxy)
        and chat.message.type == "rag_message_generator"
    ):
        message = f"{agent_names[sender.id]}.message_generator"
        chat_string += "\n" + f'{tab}    "message": {message},'
        chat_string += "\n" + tab + "},"
        return chat_string, ""
    additional_methods_string = ""
    message, method_content = get_chat_message_string(
        sender=sender,
        chat=chat,
        chat_names=chat_names,
    )
    if message and isinstance(chat.data.message, WaldiezChatMessage):
        if chat.data.message.type == "method":
            if method_content:
                additional_methods_string += "\n" + method_content
            chat_string += "\n" + f'{tab}    "message": {message},'
        elif chat.data.message.type == "string" and chat.data.message.content:
            chat_string += "\n" + f'{tab}    "message": "{message}",'
    chat_string += "\n" + tab + "},"
    return chat_string, additional_methods_string


def _get_chat_strig_start(
    connection: WaldiezAgentConnection,
    agent_names: dict[str, str],
    serializer: Callable[..., str],
    tabs: int,
    skip_cache: bool,
) -> str:
    tab = "    " * tabs
    chat = connection["chat"]
    sender = connection["source"]
    recipient = connection["target"]
    chat_args = chat.get_chat_args(for_queue=True, sender=sender)
    # chat_args = update_summary_chat_args(chat_args)
    chat_string = "{"
    chat_string += "\n" + f'{tab}    "sender": {agent_names[sender.id]},'
    chat_string += "\n" + f'{tab}    "recipient": {agent_names[recipient.id]},'
    if skip_cache is False:
        chat_string += "\n" + f'{tab}    "cache": cache,'
    # additional_methods_string = ""
    for key, value in chat_args.items():
        if isinstance(value, (dict, str)):
            chat_string += (
                "\n" + f'{tab}    "{key}": {serializer(value, tabs=tabs + 1)},'
            )
        else:
            chat_string += "\n" + f'{tab}    "{key}": {value},'
    return chat_string


def _get_initiate_chats_line(
    tab: str,
    is_async: bool,
) -> str:
    """Get the initiate chats line.

    Parameters
    ----------
    tab : str
        The tab string.
    is_async : bool
        Whether the chat is asynchronous.

    Returns
    -------
    str
        The initiate chats line.
    """
    results_is = f"{tab}results = "
    initiate = "initiate_chats"
    if is_async:
        results_is += "await "
        initiate = "a_initiate_chats"
    return results_is + initiate + "(["



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# flake8: noqa: E501
# pylint: disable=too-many-try-statements,broad-exception-caught,line-too-long

"""A Redis I/O stream for handling print and input messages."""

import json
import logging
import time
import traceback as tb
import uuid
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Optional,
    Type,
)

try:
    import redis
    import redis.asyncio as a_redis
except ImportError as error:  # pragma: no cover
    raise ImportError(
        "Redis client not installed. Please install redis-py with `pip install redis`."
    ) from error
from autogen.io import IOStream  # type: ignore
from autogen.messages import BaseMessage  # type: ignore

from .models import (
    PrintMessage,
    TextMediaContent,
    UserInputData,
    UserInputRequest,
    UserResponse,
)
from .utils import gen_id, now

if TYPE_CHECKING:
    Redis = redis.Redis[bytes]
    AsyncRedis = a_redis.Redis[bytes]
else:
    Redis = redis.Redis
    AsyncRedis = a_redis.Redis

LOG = logging.getLogger(__name__)


class RedisIOStream(IOStream):
    """Redis I/O stream."""

    redis: Redis
    task_id: str
    input_timeout: int
    on_input_request: Optional[Callable[[str, str, str], None]]
    on_input_received: Optional[Callable[[str, str], None]]
    max_stream_size: int
    output_stream: str
    input_request_channel: str
    input_response_channel: str

    def __init__(
        self,
        redis_url: str = "redis://localhost:6379/0",
        task_id: str | None = None,
        input_timeout: int = 120,
        max_stream_size: int = 1000,
        on_input_request: Optional[Callable[[str, str, str], None]] = None,
        on_input_response: Optional[Callable[[str, str], None]] = None,
        redis_connection_kwargs: dict[str, Any] | None = None,
        uploads_root: Path | str | None = None,
    ) -> None:
        """Initialize the Redis I/O stream.

        Parameters
        ----------
        task_id : str, optional
            An ID to use for the input channel and the output stream. If not provided,
            a random UUID will be generated.
        redis_url : str, optional
            The Redis URL, by default "redis://localhost:6379/0".
        input_timeout : int, optional
            The time to wait for user input in seconds, by default 180 (3 minutes).
        on_input_request : Optional[Callable[[str, str, str], None]], optional
            Callback for input request, by default None
            parameters: prompt, request_id, task_id
        on_input_response : Optional[Callable[[str, str], None]], optional
            Callback for input response, by default None.
            parameters: user_input, task_id
        redis_connection_kwargs : dict[str, Any] | None, optional
            Additional Redis connection kwargs, to be used with `redis.Redis.from_url`,
            by default None.
            See: https://redis-py.readthedocs.io/en/stable/connections.html#redis.Redis.from_url
        max_stream_size : int, optional
            The maximum number of entries per stream, by default 1000.
        uploads_root : Path | str | None, optional
            The root directory for uploads, by default None.
            If provided, it will be resolved to an absolute path.
        """
        self.redis = Redis.from_url(redis_url, **redis_connection_kwargs or {})
        self.task_id = task_id or uuid.uuid4().hex
        self.input_timeout = input_timeout
        self.on_input_request = on_input_request
        self.on_input_response = on_input_response
        self.max_stream_size = max_stream_size
        self.task_output_stream = f"task:{self.task_id}:output"
        self.input_request_channel = f"task:{self.task_id}:input_request"
        self.input_response_channel = f"task:{self.task_id}:input_response"
        self.common_output_stream = "task-output"
        self.uploads_root = (
            Path(uploads_root).resolve() if uploads_root else None
        )
        if self.uploads_root and not self.uploads_root.exists():
            self.uploads_root.mkdir(parents=True, exist_ok=True)

    def __enter__(self) -> "RedisIOStream":
        """Enable context manager usage."""
        return self

    def __exit__(
        self,
        exc_type: Type[Exception] | None,
        exc_value: Exception | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager.

        Parameters
        ----------
        exc_type : Type[Exception] | None
            The exception type.
        exc_value : Exception | None
            The exception value.
        traceback : TracebackType | None
            The traceback.
        """
        # cleanup
        RedisIOStream.cleanup_processed_task_requests(
            self.redis, self.task_id, retention_period=86400
        )
        RedisIOStream.trim_task_output_streams(self.redis)
        RedisIOStream.cleanup_processed_requests(self.redis)
        # and close the connection
        self.close()

    def close(self) -> None:
        """Close the Redis client."""
        RedisIOStream.try_do(self.redis.close)

    def _print_to_task_output(self, payload: dict[str, Any]) -> None:
        """Print message to the task output stream.

        Parameters
        ----------
        message : str
            The message to print.
        message_type : str
            The message type.
        """
        LOG.debug("Sending print message: %s", payload)
        RedisIOStream.try_do(
            self.redis.xadd,  # pyright: ignore
            self.task_output_stream,
            payload,
            maxlen=self.max_stream_size,
            approximate=True,
        )

    def _print_to_common_output(self, payload: dict[str, Any]) -> None:
        """Print message to the common output stream.

        Parameters
        ----------
        message : str
            The message to print.
        message_type : str
            The message type.
        """
        LOG.debug("Sending print message: %s", payload)
        RedisIOStream.try_do(
            self.redis.xadd,  # pyright: ignore
            self.common_output_stream,
            payload,
            maxlen=self.max_stream_size,
            approximate=True,
        )

    def _print(self, payload: dict[str, Any]) -> None:
        """Print message to Redis streams.

        Parameters
        ----------
        payload : dict[str, Any]
            The message to print.
        """
        if "id" not in payload:
            payload["id"] = gen_id()
        payload["task_id"] = self.task_id
        if "timestamp" not in payload:
            payload["timestamp"] = now()
        self._print_to_task_output(payload)
        self._print_to_common_output(payload)

    def print(self, *args: Any, **kwargs: Any) -> None:
        """Print message to Redis stream.

        Parameters
        ----------
        args : Any
            The message to print.
        kwargs : Any
            Additional keyword arguments.
        """
        print_message = PrintMessage.create(*args, **kwargs)
        payload = print_message.model_dump(mode="json")
        self._print(payload)

    def input(
        self,
        prompt: str = "",
        *,
        password: bool = False,
        request_id: str | None = None,
    ) -> str:
        """Request input via Redis Pub/Sub and wait for response.

        Parameters
        ----------
        prompt : str, optional
            The prompt message, by default "".
        password : bool, optional
            Whether input is masked, by default False.
        request_id : str, optional
            The request ID (for testing), by default None.

        Returns
        -------
        str
            The received user input, or empty string if timeout occurs.
        """
        request_id = request_id or gen_id()
        input_request = UserInputRequest(
            request_id=request_id,
            prompt=prompt,
            password=password,
        )
        payload = input_request.model_dump(mode="json")
        payload["password"] = str(password).lower()
        payload["task_id"] = self.task_id
        LOG.debug("Requesting input via Pub/Sub: %s", payload)
        self._print(payload)
        RedisIOStream.try_do(
            self.redis.publish,
            self.input_request_channel,
            json.dumps(payload),
        )
        if self.on_input_request:
            self.on_input_request(prompt, request_id, self.task_id)
        user_input = self._wait_for_input(request_id)
        if self.on_input_response:
            self.on_input_response(user_input, self.task_id)
        text_response = UserInputData(content=TextMediaContent(text=user_input))
        user_response = UserResponse(
            type="input_response",
            request_id=request_id,
            data=text_response,
        )
        payload = user_response.model_dump(mode="json")
        # no nested dicts :(
        payload["data"] = json.dumps(payload["data"])
        payload["task_id"] = self.task_id
        LOG.debug("Sending input response: %s", payload)
        self._print(payload)
        return user_input

    def send(self, message: BaseMessage) -> None:
        """Send a structured message to Redis.

        Parameters
        ----------
        message : dict[str, Any]
            The message to send.
        """
        try:
            message_dump = message.model_dump(mode="json")
        except Exception as e:  # pragma: no cover
            message_dump = {
                "type": "error",
                "error": str(e),
            }
        message_type = message_dump.get("type", None)
        if not message_type:
            message_type = message.__class__.__name__
        self._print(
            {
                "data": json.dumps(message_dump),
                "type": message_type,
            }
        )

    def _wait_for_input(self, input_request_id: str) -> str:
        """Wait for user input.

        Parameters
        ----------
        input_request_id : str
            The request ID.

        Returns
        -------
        str
            The user input.
        """
        lock_key = f"lock:{self.task_id}"
        start_time = time.time()

        pubsub = self.redis.pubsub()
        pubsub.subscribe(self.input_response_channel)
        try:
            while (time.time() - start_time) <= self.input_timeout:
                message = pubsub.get_message(ignore_subscribe_messages=True)
                if not message:
                    time.sleep(0.1)
                    continue
                LOG.debug("Received message: %s", message)
                response = self.parse_pubsub_input(message)
                if not response or response.request_id != input_request_id:
                    continue

                if self._acquire_lock(lock_key):  # pragma: no branch
                    try:
                        if self._is_request_processed(response.request_id):
                            continue

                        self._mark_request_processed(response.request_id)
                        return self._get_user_input(response)
                    finally:
                        self._release_lock(lock_key)
        except BaseException:  # pragma: no cover
            LOG.error("Error in _wait_for_input: %s", tb.format_exc())
        finally:
            pubsub.unsubscribe(self.input_response_channel)

        LOG.warning(
            "No input received for %ds on task %s, assuming empty string",
            self.input_timeout,
            self.task_id,
        )
        return ""

    # pylint:disable=no-self-use
    def _get_user_input(self, response: UserResponse) -> str:
        """Get user input from the response.

        Parameters
        ----------
        response : UserResponse
            The user response.

        Returns
        -------
        str
            The user input.
        """
        if not response.data:
            return ""
        if isinstance(
            response.data, str
        ):  # pragma: no cover should be structured
            return response.data
        return response.to_string(
            uploads_root=self.uploads_root,
            base_name=response.request_id,
        )

    def _acquire_lock(self, lock_key: str, lock_expiry: int = 10) -> bool:
        """Try to acquire a lock, returns True if acquired, False otherwise."""
        try:
            return (
                self.redis.set(lock_key, "locked", ex=lock_expiry, nx=True)
                is True
            )
        except redis.RedisError as e:  # pragma: no cover
            LOG.error("Redis error on acquire lock: %s", e)
            return False
        except BaseException as e:  # pragma: no cover
            LOG.error("Error on acquire lock: %s", e)
            return False

    def _release_lock(self, lock_key: str) -> None:
        """Release a lock."""
        RedisIOStream.try_do(self.redis.delete, lock_key)

    def _is_request_processed(self, request_id: str) -> bool:
        """Check if a request is processed for a task."""
        return RedisIOStream.is_request_processed(
            self.redis, task_id=self.task_id, request_id=request_id
        )

    def _mark_request_processed(self, request_id: str) -> None:
        """Mark a request as processed for a task."""
        RedisIOStream.try_do(
            self.redis.zadd,
            f"processed_requests:{self.task_id}",
            {request_id: int(time.time() * 1_000_000)},
        )

    @staticmethod
    def _extract_message_data(data: Any) -> Optional[dict[str, Any]]:
        """Extract and parse the message data field."""
        message_data = data

        # Handle string-encoded JSON
        if isinstance(message_data, str):
            try:
                message_data = json.loads(message_data)
            except json.JSONDecodeError:
                LOG.error("Invalid JSON in message data: %s", message_data)
                return None

        # Validate data type
        if not isinstance(message_data, dict):  # pragma: no cover
            LOG.error("Invalid message data format: %s", message_data)
            return None

        return message_data  # pyright: ignore

    @staticmethod
    def _message_has_required_fields(message_data: dict[str, Any]) -> bool:
        """Check if message data contains required fields."""
        if "request_id" not in message_data:
            LOG.error("Missing 'request_id' in message data: %s", message_data)
            return False

        return True

    @staticmethod
    def _process_nested_data(
        message_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Process nested JSON data if present."""
        # Create a copy to avoid modifying the original
        processed_data = message_data.copy()

        # Handle nested JSON in 'data' field
        if "data" in processed_data and isinstance(
            processed_data["data"], str
        ):  # pragma: no branch
            try:
                processed_data["data"] = json.loads(processed_data["data"])
            except json.JSONDecodeError:
                LOG.error(
                    "Invalid JSON in nested data field: %s", processed_data
                )
                return None

        return processed_data

    @staticmethod
    def _create_user_response(
        message_data: dict[str, Any],
    ) -> Optional["UserResponse"]:
        """Create UserResponse object from validated data."""
        try:
            return UserResponse.model_validate(message_data)
        except Exception as e:
            LOG.error(
                "Error parsing user input response: %s - %s",
                message_data,
                str(e),
            )
            return None

    def parse_pubsub_input(
        self,
        message: dict[str, Any] | None,
    ) -> UserResponse | None:
        """Extract request ID and user input from a message.

        Parameters
        ----------
        message : dict[str, Any]
            The message to parse.

        Returns
        -------
        UserResponse
            The parsed user response.
        """
        if not isinstance(message, dict) or "data" not in message:
            LOG.error("Invalid message format or missing 'data': %s", message)
            return None
        message_data = self._extract_message_data(message["data"])
        if message_data is None:  # pragma: no cover
            return None

        if not self._message_has_required_fields(
            message_data
        ):  # pragma: no cover
            return None

        processed_data = self._process_nested_data(message_data)
        if processed_data is None:  # pragma: no cover
            return None

        return self._create_user_response(processed_data)

    @staticmethod
    def try_do(func: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Try to execute.

        Just to avoid duplicate try/except blocks.
        To only be used if no return value is expected.
        And if we no't need to re-raise the exception.
        Otherwise, we normally try/except at the call site.

        Parameters
        ----------
        func : Callable[..., Any]
            The function to call.
        args : Any
            The function's positional arguments.
        kwargs : Any
            The function's keyword arguments.
        """
        try:
            func(*args, **kwargs)
        except BaseException:  # pragma: no cover
            LOG.error("Error on try_do:")
            LOG.error(tb.format_exc())

    @staticmethod
    async def a_try_do(
        func: Callable[..., Awaitable[Any]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Async version of try_do.

        Parameters
        ----------
        func : Awaitable[Any]
            The async function to call.
        args : Any
            The positional arguments.
        kwargs : Any
            The keyword arguments.
        """
        try:
            await func(*args, **kwargs)
        except BaseException:  # pragma: no cover
            LOG.error("Error on a_try_do:")
            LOG.error(tb.format_exc())

    @staticmethod
    def is_request_processed(
        redis_client: Redis,
        task_id: str,
        request_id: str,
    ) -> bool:
        """Check if a request is processed for a task.

        Parameters
        ----------
        redis_client : Redis
            The async Redis client to use.
        task_id : str
            The task ID.
        request_id : str
            The request ID.

        Returns
        -------
        bool
            True if the request is processed, False otherwise.
        """
        try:
            return (
                redis_client.zscore(f"processed_requests:{task_id}", request_id)
                is not None
            )
        except BaseException as e:  # pragma: no cover
            LOG.error("Error on check request processed: %s", e)
            return False

    @staticmethod
    async def a_is_request_processed(
        redis_client: AsyncRedis,
        task_id: str,
        request_id: str,
    ) -> bool:
        """Async version of is_request_processed.

        Parameters
        ----------
        redis_client : AsyncRedis
            The async Redis client to use.
        task_id : str
            The task ID.
        request_id : str
            The request ID.

        Returns
        -------
        bool
            True if the request is processed, False otherwise.
        """
        try:
            return (
                await redis_client.zscore(
                    f"processed_requests:{task_id}", request_id
                )
                is not None
            )
        except BaseException as e:  # pragma: no cover
            LOG.error("Error on check request processed: %s", e)
            return False

    # other static methods for cleanup
    # to be used externally (like in periodic tasks) if needed
    # or after task completion
    @staticmethod
    def cleanup_processed_task_requests(
        redis_client: Redis,
        task_id: str,
        retention_period: int = 86400,
    ) -> None:
        """Cleanup old processed request logs.

        Parameters
        ----------
        redis_client : Redis
            The Redis client.
        task_id : str
            The task ID.
        retention_period : int, optional
            The retention period in seconds, by default 86400.
        """
        key = f"processed_requests:{task_id}"
        cutoff_time = int(time.time()) - retention_period
        RedisIOStream.try_do(redis_client.zremrangebyscore, key, 0, cutoff_time)

    @staticmethod
    def cleanup_processed_requests(
        redis_client: Redis, retention_period: int = 86400
    ) -> None:
        """Cleanup stale processed requests.

        Parameters
        ----------
        redis_client : Redis
            The Redis client.
        retention_period : int, optional
            The retention period in seconds
        """
        cutoff_time = int(time.time()) - retention_period
        for key in redis_client.scan_iter("processed_requests:*", count=100):
            RedisIOStream.try_do(
                redis_client.zremrangebyscore, key, 0, cutoff_time
            )

    @staticmethod
    def trim_task_output_streams(
        redis_client: Redis, maxlen: int = 1000, approximate: bool = True
    ) -> None:
        """Trim task output streams to a max length.

        Parameters
        ----------
        redis_client : Redis
            The Redis client.
        maxlen : int
            The maximum number of entries per stream.
        approximate : bool
            Whether to use approximate trimming (more efficient).
        """
        for key in redis_client.scan_iter("task:*:output", count=100):
            RedisIOStream.try_do(
                redis_client.xtrim,  # pyright: ignore
                key,
                maxlen=maxlen,
                approximate=approximate,
            )

    @staticmethod
    async def a_cleanup_processed_task_requests(
        redis_client: AsyncRedis, task_id: str, retention_period: int = 86400
    ) -> None:
        """Async version of cleanup task processed requests.

        Parameters
        ----------
        redis_client : AsyncRedis
            The Redis client.
        task_id : str
            The task ID.
        retention_period : int, optional
            The retention period in seconds, by default 86400.
        """
        key = f"processed_requests:{task_id}"
        cutoff_time = int(time.time()) - retention_period
        await RedisIOStream.a_try_do(
            redis_client.zremrangebyscore, key, 0, cutoff_time
        )

    @staticmethod
    async def a_cleanup_processed_requests(
        redis_client: AsyncRedis,
        retention_period: int = 86400,
    ) -> None:
        """Async version of cleanup stale processed requests.

        Parameters
        ----------
        redis_client : Redis
            The Redis client.
        retention_period : int, optional
            The retention period in seconds, by default 86400.
        """
        cutoff_time = int(time.time()) - retention_period
        async for key in redis_client.scan_iter(
            "processed_requests:*", count=100
        ):
            await RedisIOStream.a_try_do(
                redis_client.zremrangebyscore, key, 0, cutoff_time
            )

    @staticmethod
    async def a_trim_task_output_streams(
        redis_client: AsyncRedis,
        maxlen: int = 1000,
        approximate: bool = True,
        scan_count: int = 100,
    ) -> None:
        """Trim task output Redis streams to a max length.

        Parameters
        ----------
        redis_client : AsyncRedis
            The Redis client.
        maxlen : int
            The maximum number of entries per stream.
        approximate : bool
            Whether to use approximate trimming (more efficient).
        scan_count : int
            The number of keys to scan per iteration.
        """
        trimmed_count = 0

        async for key in redis_client.scan_iter(
            "task:*:output", count=scan_count
        ):  # pragma: no branch
            before = await redis_client.xlen(key)
            await RedisIOStream.a_try_do(
                redis_client.xtrim,  # pyright: ignore
                key,
                maxlen=maxlen,
                approximate=approximate,
            )
            after = await redis_client.xlen(key)
            if before > after:  # pragma: no branch
                trimmed = before - after
                trimmed_count += trimmed
                LOG.debug("Trimmed %d entries from %s", trimmed, key)

        LOG.info("Total trimmed entries: %d", trimmed_count)
        # we might also want to use prometheus metrics here,
        # to check and fine-tune the (maxlen, scan_count) parameters
        # e.g.:
        # from prometheus_client import Counter

        # trimmed_entries = Counter("redis_stream_trimmed_entries_total", "Total trimmed entries from Redis streams")
        # trimmed_streams = Counter("redis_stream_trimmed_streams_total", "Total number of trimmed Redis streams")
        # ...
        # if before > after:
        #     trimmed_entries.inc(before - after)
        #     trimmed_streams.inc()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pyright: reportCallIssue=false
"""Waldiez flow model."""

from functools import cached_property
from typing import Optional

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ..agents import WaldiezAgent, WaldiezGroupManager
from ..chat import WaldiezChat
from ..common import WaldiezBase, get_id, get_waldiez_version, now
from .connection import WaldiezAgentConnection
from .flow_data import WaldiezFlowData
from .naming import WaldiezUniqueNames, ensure_unique_names


class WaldiezFlow(WaldiezBase):
    """Flow data class.

    Attributes
    ----------
    id : str
        The ID of the flow.
    type : Literal["flow"], optional
        The type of the "node" in a graph: "flow".
    name : str
        The name of the flow.
    description : str
        The description of the flow.
    tags : list[str]
        The tags of the flow.
    requirements : list[str]
        The requirements of the flow.
    storage_id : str
        The storage ID of the flow (ignored, UI related).
    created_at : str
        The date and time when the flow was created.
    updated_at : str
        The date and time when the flow was last updated.
    data : WaldiezFlowData
        The data of the flow. See `WaldiezFlowData`.
    """

    id: Annotated[
        str,
        Field(
            description="The ID of the flow",
            title="ID",
            default_factory=get_id,
        ),
    ]
    version: Annotated[
        str,
        Field(
            default_factory=get_waldiez_version,
            description="The version waldiez that was used to create the flow",
            title="Version",
        ),
    ] = "0.0.0"  # default to 0.0.0 if not found
    type: Annotated[
        Literal["flow"],
        Field(
            "flow",
            description="The type of the 'node' in a graph",
            title="Type",
        ),
    ]
    name: Annotated[
        str,
        Field(
            ...,
            description="The name of the flow",
            title="Name",
        ),
    ]
    description: Annotated[
        str,
        Field(
            ...,
            description="The description of the flow",
            title="Description",
        ),
    ]
    tags: Annotated[
        list[str],
        Field(
            description="The tags of the flow",
            title="Tags",
            default_factory=list,
        ),
    ]
    requirements: Annotated[
        list[str],
        Field(
            description="The requirements of the flow",
            title="Requirements",
            default_factory=list,
        ),
    ]
    data: Annotated[
        WaldiezFlowData,
        Field(
            ...,
            description="The data of the flow",
            title="Data",
        ),
    ]
    storage_id: Annotated[
        str,
        Field(
            default_factory=get_id,
            description="The storage ID of the flow (ignored, UI related)",
            title="Storage ID",
            alias="storageId",
        ),
    ]
    created_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Created At",
            description="The date and time when the flow was created.",
        ),
    ]
    updated_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Updated At",
            description="The date and time when the flow was last updated.",
        ),
    ]
    _ordered_flow: Optional[list[WaldiezAgentConnection]] = None
    _single_agent_mode: bool = False
    _is_group_chat: bool = False

    @property
    def is_async(self) -> bool:
        """Check if the flow is asynchronous.

        Returns
        -------
        bool
            True if the flow is asynchronous, False otherwise.
        """
        return self.data.is_async

    @property
    def cache_seed(self) -> Optional[int]:
        """Check if the flow has caching disabled.

        Returns
        -------
        bool
            True if the flow has caching disabled, False otherwise.
        """
        return self.data.cache_seed

    @property
    def is_single_agent_mode(self) -> bool:
        """Check if the flow is in single agent mode.

        Returns
        -------
        bool
            True if the flow is in single agent mode, False otherwise.
        """
        return self._single_agent_mode

    @property
    def ordered_flow(
        self,
    ) -> list[WaldiezAgentConnection]:
        """Get the ordered flow."""
        if self._ordered_flow is None:
            self._ordered_flow = self._get_flow_order()
        return self._ordered_flow

    @cached_property
    def unique_names(self) -> WaldiezUniqueNames:
        """Get the unique names for the flow.

        Returns
        -------
        WaldiezUniqueNames
            The unique names for the flow.
        """
        return ensure_unique_names(
            flow_id=self.id,
            flow_name=self.name,
            flow_agents=self.data.agents.members,
            flow_chats=self.data.chats,
            flow_tools=self.data.tools,
            flow_models=self.data.models,
        )

    def get_agent_by_id(self, agent_id: str) -> WaldiezAgent:
        """Get the agent by ID.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.

        Returns
        -------
        WaldiezAgent
            The agent.

        Raises
        ------
        ValueError
            If the agent with the given ID is not found.
        """
        for agent in self.data.agents.members:
            if agent.id == agent_id:
                return agent
        raise ValueError(f"Agent with ID {agent_id} not found.")

    @classmethod
    def default(cls) -> "WaldiezFlow":
        """Get the default flow.

        Returns
        -------
        WaldiezFlow
            The default flow.
        """
        an_id = get_id()
        return cls(
            id=an_id,
            storage_id=an_id,
            created_at=now(),
            updated_at=now(),
            type="flow",
            name="Default Flow",
            description="Default Flow",
            tags=[],
            requirements=[],
            data=WaldiezFlowData.default(),
        )

    def get_group_members(self, group_id: str) -> list[WaldiezAgent]:
        """Get the group members.

        Parameters
        ----------
        group_id : str
            The ID of the group.

        Returns
        -------
        list[WaldiezAgent]
            The list of group members.
        """
        return [
            agent
            for agent in self.data.agents.members
            if agent.data.parent_id == group_id
        ]

    def get_agent_connections(
        self, agent_id: str, all_chats: bool = True
    ) -> list[str]:
        """Get the agent connections.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.
        all_chats : bool, optional
            If True, get the connections from all the chats, otherwise
            get the connections from the ordered flow (main chat flow).

        Returns
        -------
        list[str]
            The list of agent ids that the agent with the given ID connects to.
        """
        connections: list[str] = []
        if all_chats:
            for chat in self.data.chats:
                if chat.source == agent_id:
                    connections.append(chat.target)
                if chat.target == agent_id:
                    connections.append(chat.source)
        else:
            for entry in self.ordered_flow:
                source = entry["source"]
                target = entry["target"]
                if source.id == agent_id:
                    connections.append(target.id)
                if target.id == agent_id:
                    connections.append(source.id)
        return connections

    def get_group_chat_members(
        self, group_manager_id: str
    ) -> list[WaldiezAgent]:
        """Get the group chat members.

        Parameters
        ----------
        group_manager_id : str
            The ID of the group manager.

        Returns
        -------
        list[WaldiezAgent]
            The list of group chat members.
        """
        agent = self.get_agent_by_id(group_manager_id)
        if not agent.is_group_manager or not isinstance(
            agent, WaldiezGroupManager
        ):
            return []
        members = [
            agent
            for agent in self.data.agents.members
            if agent.data.parent_id == group_manager_id
        ]
        if agent.data.speakers.selection_method != "round_robin":
            return members
        ordered_ids = agent.get_speakers_order()
        if not ordered_ids:
            return members
        members_dict = {member.id: member for member in members}
        ordered_members = [
            members_dict[member_id]
            for member_id in ordered_ids
            if member_id in members_dict
        ]
        return ordered_members

    def _get_flow_order(
        self,
    ) -> list[WaldiezAgentConnection]:
        """Get the ordered flow."""
        if self._is_group_chat:
            return self._get_group_chat_flow()
        # in the chats, there is the 'order' field, we use this,
        # we only keep the ones with order >=0
        # and sort them by this property
        ordered_flow: list[WaldiezAgentConnection] = []
        for chat in self.data.chats:
            if chat.data.order < 0:
                continue
            source = self.get_agent_by_id(chat.source)
            target = self.get_agent_by_id(chat.target)
            ordered_flow.append(
                {
                    "source": source,
                    "target": target,
                    "chat": chat,
                }
            )
            # ordered_flow.append((chat, source, target))
        if not ordered_flow:
            if len(self.data.chats) == 1:
                chat = self.data.chats[0]
                source = self.get_agent_by_id(chat.source)
                target = self.get_agent_by_id(chat.target)
                ordered_flow.append(
                    {
                        "source": source,
                        "target": target,
                        "chat": chat,
                    }
                )
        return ordered_flow

    def get_root_group_manager(self) -> WaldiezGroupManager:
        """Get the root group manager.

        Returns
        -------
        WaldiezGroupManager
            The root group manager.

        Raises
        ------
        ValueError
            If no group manager is found.
        """
        for agent in self.data.agents.groupManagerAgents:
            if agent.data.parent_id is None:
                return agent
        raise ValueError("No group manager found.")

    def _get_group_chat_flow(
        self,
    ) -> list[WaldiezAgentConnection]:
        """Get the ordered flow for group chat.

        Returns
        -------
        list[tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
            The ordered flow for group chat.
        """
        # in a group chat there is no "order", the group manager
        # handles the conversation (using the group "pattern")
        # the only thin to check is if there is a user agent that connects
        # to the group manager agent (so that would be the first chat)
        # if found, we must then check if the message from the user
        # to the group manager is:
        #  - "text" or "none" => no need to create a group manager on ag2
        #  - "function/method" => create a group manager and a group chat on ag2
        # in the first case, the chat would be:
        # result, context, last_agent = initiate_group_chat(
        #     pattern=pattern,
        #     messages=...,
        #     max_rounds=10
        # )
        # in the second case, the chat would be:
        # user.initiate_chat(manager, ...)
        user_agent: Optional[WaldiezAgent] = None
        to_root_manager: Optional[WaldiezChat] = None
        root_manager: WaldiezGroupManager = self.get_root_group_manager()
        for chat in self.data.chats:
            if chat.target == root_manager.id:
                # check if the source is a user agent
                source = self.get_agent_by_id(chat.source)
                if source.is_user:
                    user_agent = source
                    to_root_manager = chat
                    break
        if not to_root_manager or not user_agent:
            return []
        return [
            {
                "source": user_agent,
                "target": root_manager,
                "chat": to_root_manager,
            }
        ]

    def _validate_agent_connections(self) -> None:
        for agent in self.data.agents.members:
            if agent.is_group_member:
                # group members are allowed
                # to not connect to any other node
                # the group manager will take care of
                # the agent/speaker connections
                continue
            if not any(
                agent.id in (chat.source, chat.target)
                for chat in self.data.chats
            ):
                raise ValueError(
                    f"Agent {agent.id} ({agent.name}) "
                    "does not connect to any other node."
                )

    @model_validator(mode="after")
    def validate_flow(self) -> Self:
        """Flow validation.

        - unique node ids
        - there are at least two agents
            - (or a single agent but not a group manager)
        - all the agents connect to at least one other agent
        - all the linked agent tools are found in the flow
        - all the linked agent models are found in the flow
        - all the managers have at least one member in the chat group
        - the ordered flow (chats with position >=0) is not empty
        - all agents' code execution config functions exist in the flow tools
        - if group chat flow, there is at least one group manager agent
        - if group chat flow, there is an initial group member agent

        Returns
        -------
        WaldiezFlow
            The validated flow.

        Raises
        ------
        ValueError
            If the ordered flow is empty.
            If the model IDs are not unique.
            If the tool IDs are not unique.
            If the agents do not connect to any other node.
            If the manager's group chat has no members.
        """
        all_members = list(self.data.agents.members)
        all_chats = list(self.data.chats)
        for agent in all_members:
            agent.gather_nested_chats(
                all_agents=all_members, all_chats=all_chats
            )
            agent.gather_handoffs(all_agents=all_members, all_chats=all_chats)
        self._validate_group_chat(all_members)
        if len(all_members) == 1:
            return self._validate_single_agent_mode(all_members[0])
        ordered_flow = self.ordered_flow  # could be empty (if group chat)
        if not ordered_flow and self._ordered_flow is None:
            raise ValueError("The ordered flow is empty.")
        model_ids = self._validate_flow_models()
        tools_ids = self._validate_flow_tools()
        self.data.agents.validate_flow(model_ids, tools_ids)
        self._validate_agent_connections()
        return self

    def _validate_flow_models(self) -> list[str]:
        """Validate the flow models.

        Returns
        -------
        list[str]
            The list of model IDs.

        Raises
        ------
        ValueError
            If the model IDs are not unique.
        """
        model_ids = [model.id for model in self.data.models]
        if len(model_ids) != len(set(model_ids)):
            raise ValueError("Model IDs must be unique.")
        return model_ids

    def _validate_flow_tools(self) -> list[str]:
        """Validate the flow tools.

        Returns
        -------
        list[str]
            The list of tool IDs.

        Raises
        ------
        ValueError
            If the tool IDs are not unique.
        """
        tool_ids = [tool.id for tool in self.data.tools]
        if len(tool_ids) != len(set(tool_ids)):
            raise ValueError("Tool IDs must be unique.")
        return tool_ids

    def _validate_single_agent_mode(self, member: WaldiezAgent) -> Self:
        """Flow validation for single agent mode.

        Parameters
        ----------
        member : WaldiezAgent
            The only agent in the flow

        Returns
        -------
        WaldiezFlow
            The validated flow.

        Raises
        ------
        ValueError
            - If the only agent is a group manager.
            - If the model IDs are not unique.
            - If the tool IDs are not unique.
        """
        if member.is_group_manager:
            raise ValueError(
                "In single agent mode, the agent must not be a group manager."
            )
        model_ids = self._validate_flow_models()
        tools_ids = self._validate_flow_tools()
        self.data.agents.validate_flow(model_ids, tools_ids)
        self._single_agent_mode = True
        return self

    def _validate_group_manager(
        self, group_manager: WaldiezGroupManager, all_member_ids: list[str]
    ) -> None:
        """Validate the group manager agents.

        Raises
        ------
        ValueError
            If there are no group manager agents.
        """
        if not group_manager.data.initial_agent_id:
            raise ValueError(
                "The flow is a group chat but the group manager agent "
                f"{group_manager.id} has no initial agent ID."
            )
        if group_manager.data.initial_agent_id not in all_member_ids:
            raise ValueError(
                "The flow is a group chat but the initial agent ID "
                f"{group_manager.data.initial_agent_id} is not in the flow."
            )
        group_members = self.get_group_members(group_manager.id)
        if not group_members:
            raise ValueError(
                "The flow is a group chat but the group manager agent "
                f"{group_manager.id} has no members in the group."
            )
        group_manager.set_speakers_order(
            [member.id for member in group_members]
        )

    def _validate_group_chat(self, all_members: list[WaldiezAgent]) -> None:
        """Check if the flow is a group chat and validate it.

        Raises
        ------
        ValueError
            If the flow is a group chat and there is no group manager agent,
            if the group has no members,
            or if the group has no initial member agent.
        """
        if not self.data.agents.groupManagerAgents:
            # no group manager agents, not a group chat
            return
        self._is_group_chat = True
        if not any(agent.is_group_member for agent in self.data.agents.members):
            raise ValueError(
                "The flow is a group chat but has no members in the group."
            )
        # check if the group manager agents are the flow
        group_manager_ids = [
            agent.id for agent in self.data.agents.groupManagerAgents
        ]
        all_member_ids = [agent.id for agent in all_members]
        if not all(
            group_manager_id in all_member_ids
            for group_manager_id in group_manager_ids
        ):
            raise ValueError(
                "The flow is a group chat but not all group manager agents are "
                "in the flow."
            )
        # check the initial_agent_id for each group
        for group_manager in self.data.agents.groupManagerAgents:
            self._validate_group_manager(group_manager, all_member_ids)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods,no-self-use
"""Standard agent extras module."""

from dataclasses import dataclass
from typing import Optional

from ...enums import AgentPosition, ExportPosition
from ...result import ExportResult
from ...types import ImportStatement, InstanceArgument
from ..base import BaseExtras


@dataclass
class TerminationConfig:
    """Termination configuration."""

    termination_arg: str = "None"
    before_content: str = ""

    def has_content(self) -> bool:
        """Check if there's any termination content.

        Returns
        -------
        bool
            True if there's any termination configuration.
        """
        return bool(self.before_content.strip())


@dataclass
class CodeExecutionConfig:
    """Code execution configuration."""

    executor_content: str = ""
    executor_argument: str = "False"
    executor_import: ImportStatement | None = None

    def has_content(self) -> bool:
        """Check if there's any code execution content.

        Returns
        -------
        bool
            True if there's any code execution configuration.
        """
        return bool(
            self.executor_content.strip()
            or self.executor_argument != "False"
            or self.executor_import
        )


@dataclass
class SystemMessageConfig:
    """System message configuration."""

    # in case we support later custom methods for system messages
    before_agent_conent: str = ""
    system_message_arg: str = ""

    def has_content(self) -> bool:
        """Check if there's any system message content.

        Returns
        -------
        bool
            True if there's any system message configuration.
        """
        return bool(str(self.system_message_arg).strip())


@dataclass
class StandardExtras(BaseExtras):
    """Extras for standard agents (UserProxy, Assistant, etc.)."""

    code_execution_config: Optional[CodeExecutionConfig] = None
    termination_config: Optional[TerminationConfig] = None
    system_message_config: Optional[SystemMessageConfig] = None

    def set_code_execution(self, config: CodeExecutionConfig) -> None:
        """Set code execution configuration.

        Parameters
        ----------
        config : CodeExecutionConfig
            The code execution configuration.
        """
        self.code_execution_config = config
        if config.executor_import:
            self.add_import(config.executor_import)
        # we either add the executor content here or
        # in _contribute_specific_content below
        if config.executor_content:
            self.append_before_agent(config.executor_content)

    def set_termination_config(self, config: TerminationConfig) -> None:
        """Set termination configuration.

        Parameters
        ----------
        config : TerminationConfig
            The termination configuration.
        """
        self.termination_config = config
        if config.before_content:
            self.append_before_agent(config.before_content)

    def set_system_message_config(self, config: SystemMessageConfig) -> None:
        """Set system message configuration.

        Parameters
        ----------
        config : SystemMessageConfig
            The system message configuration.
        """
        self.system_message_config = config
        if config.before_agent_conent:
            self.append_before_agent(config.before_agent_conent)

    def get_code_execution_arg(self) -> InstanceArgument:
        """Get the code execution argument string.

        Returns
        -------
        InstanceArgument
            The code execution argument.
        """
        if (
            self.code_execution_config
            and self.code_execution_config.executor_argument
        ):
            argunent = self.code_execution_config.executor_argument
        else:
            argunent = "False"
        return InstanceArgument(
            instance_id=self.instance_id,
            name="code_execution_config",
            value=argunent,
            tabs=1,
        )

    def get_termination_arg(self) -> InstanceArgument:
        """Get the termination argument.

        Returns
        -------
        InstanceArgument
            The termination argument.
        """
        if self.termination_config and self.termination_config.termination_arg:
            argument = self.termination_config.termination_arg
        else:
            argument = "None"
        comment: str | None = None
        if argument == "None":
            comment = "pyright: ignore"
        return InstanceArgument(
            instance_id=self.instance_id,
            name="is_termination_msg",
            value=argument,
            comment=comment,
            tabs=1,
        )

    def get_system_message_arg(self) -> InstanceArgument:
        """Get the system message argument.

        Returns
        -------
        InstanceArgument
            The system message argument.
        """
        if (
            self.system_message_config
            and self.system_message_config.system_message_arg
        ):
            argument = self.system_message_config.system_message_arg
        else:
            argument = ""
        return InstanceArgument(
            instance_id=self.instance_id,
            name="system_message",
            value=argument,
            skip_if_empty_string=True,
            with_new_line_before=True,
            tabs=1,
        )

    def has_specific_content(self) -> bool:
        """Check for standard agent specific content.

        Returns
        -------
        bool
            True if there's standard agent specific content.
        """
        return bool(
            self.extra_args
            or (
                self.code_execution_config
                and self.code_execution_config.has_content()
            )
            or (
                self.termination_config
                and self.termination_config.has_content()
            )
            or (
                self.system_message_config
                and self.system_message_config.has_content()
            )
        )

    def _contribute_specific_content(self, result: ExportResult) -> None:
        """Contribute standard agent specific content.

        Parameters
        ----------
        result : ExportResult
            The export result to contribute to.
        """
        if self.extra_args:
            result.add_content(
                self.get_extra_args_content(),
                position=ExportPosition.AGENTS,
                agent_position=AgentPosition.AS_ARGUMENT,
                agent_id=self.instance_id,
            )
        # content added above
        # alternatively, we could add it here
        # (but we should pay attention with the order)
        # if (
        #     self.code_execution_config
        #     and self.code_execution_config.has_content()
        # ):
        #     result.add_content(
        #         self.code_execution_config.executor_content,
        #         position=ExportPosition.AGENTS,
        #         agent_position=AgentPosition.BEFORE,
        #         order=ContentOrder.EARLY_SETUP,
        #         agent_id=self.instance_id,
        #     )
        #     if self.code_execution_config.executor_import:
        #    \statement = self.code_execution_config.executor_import.statement
        #         position = self.code_execution_config.executor_import.position
        #         result.add_import(
        #             statement=statement,
        #             position=position,
        #         )
        # if self.termination_config and self.termination_config.has_content():
        #     result.add_content(
        #         self.termination_config.before_content,
        #         position=ExportPosition.AGENTS,
        #         agent_position=AgentPosition.BEFORE,
        #         agent_id=self.instance_id,
        #     )
        # for extra_import in self.extra_imports:
        #     result.add_import(
        #         statement=extra_import.statement,
        #         position=extra_import.position,
        #     )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: skip-file
# isort: skip_file
# flake8: noqa: E501,E402
"""Waldiez-studio extra typer commands for CLI."""

from typing import Any, Callable

import typer
from typer.models import CommandInfo

_have_studio = False
studio_app: Callable[..., Any] | None = None

# noinspection PyBroadException
# pylint: disable=broad-exception-caught
try:
    from waldiez_studio.cli import run  # type: ignore[unused-ignore, import-not-found, import-untyped]

    studio_app = run  # pyright: ignore

    _have_studio = True
except BaseException:
    pass


def add_studio_cli(app: typer.Typer) -> None:
    """Add studio command to the app if available.

    Parameters
    ----------
    app : typer.Typer
        The Typer app to add the studio command to.
    """
    if _have_studio:
        app.registered_commands.append(
            CommandInfo(name="studio", callback=studio_app)
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
# pylint: disable=line-too-long
"""WebSocket IOStream implementation for AsyncIO."""

import asyncio
from typing import Any, Protocol

HAS_WS_LIB = False

try:
    from starlette.websockets import WebSocket  # type: ignore[unused-ignore, unused-import, import-not-found, import-untyped]  # noqa

    HAS_WS_LIB = True  # pyright: ignore
except ImportError:
    pass

try:
    import websockets  # type: ignore[unused-ignore, unused-import, import-not-found, import-untyped]  # noqa

    HAS_WS_LIB = True  # pyright: ignore
except ImportError:  # pragma: no cover
    pass


class WebSocketConnection(Protocol):
    """Protocol for WebSocket connections."""

    async def send_message(self, message: str) -> None:
        """Send a message over the WebSocket connection.

        Parameters
        ----------
        message : str
            The message to send.
        """

    async def receive_message(
        self,
        timeout: float = 120,
    ) -> str:  # pyright: ignore
        """Receive a message from the WebSocket connection.

        Parameters
        ----------
        timeout : float, optional
            The timeout for receiving the message. Defaults to 120 seconds.
            If the timeout is reached, an empty string is returned.
            If an error occurs, an empty string is returned.

        Returns
        -------
        str
            The received message.
        """


class WebSocketsAdapter:
    """Adapter for websockets library connections."""

    def __init__(self, websocket: "websockets.ServerConnection"):
        """Initialize the adapter.

        Parameters
        ----------
        websocket : ServerConnection
            The websockets library connection.
        """
        self.websocket = websocket

    async def send_message(self, message: str) -> None:
        """Send a message using websockets library.

        Parameters
        ----------
        message : str
            The message to send.
        """
        await self.websocket.send(message)

    async def receive_message(self, timeout: float = 120) -> str:
        """Receive a message using websockets library.

        Parameters
        ----------
        timeout : float, optional
            The timeout for receiving the message. Defaults to 120 seconds.

        Returns
        -------
        str
            The received message, decoded as a string.
        """
        # pylint: disable=too-many-try-statements
        try:
            response = await asyncio.wait_for(
                self.websocket.recv(), timeout=timeout
            )
            if isinstance(response, bytes):
                return response.decode("utf-8")
            return response if isinstance(response, str) else str(response)
        except asyncio.TimeoutError:
            return ""
        except BaseException:  # pylint: disable=broad-exception-caught
            # Handle other exceptions that may occur during receive
            return ""


class StarletteAdapter:
    """Adapter for Starlette/FastAPI WebSocket connections."""

    def __init__(self, websocket: "WebSocket"):
        """Initialize the adapter.

        Parameters
        ----------
        websocket : WebSocket
            The Starlette/FastAPI WebSocket connection.
        """
        self.websocket = websocket

    async def send_message(self, message: str) -> None:
        """Send a message using Starlette WebSocket.

        Parameters
        ----------
        message : str
            The message to send.
        """
        await self.websocket.send_text(message)

    async def receive_message(self, timeout: float = 120) -> str:
        """Receive a message using Starlette WebSocket.

        Parameters
        ----------
        timeout : float, optional
            The timeout for receiving the message. Defaults to 120 seconds.

        Returns
        -------
        str
            The received message, decoded as a string.
        """
        try:
            return await asyncio.wait_for(
                self.websocket.receive_text(), timeout=timeout
            )
        except asyncio.TimeoutError:
            return ""
        except BaseException:  # pylint: disable=broad-exception-caught
            # Handle other exceptions that may occur during receive
            return ""


def create_websocket_adapter(websocket: Any) -> WebSocketConnection:
    """Create an appropriate adapter for the given WebSocket connection.

    Parameters
    ----------
    websocket : Any
        The WebSocket connection (websockets or Starlette).

    Returns
    -------
    WebSocketConnection
        An adapter that implements the WebSocketConnection protocol.

    Raises
    ------
    ValueError
        If the WebSocket type is not supported.
    """
    # Check for Starlette/FastAPI WebSocket
    if hasattr(websocket, "send_text") and hasattr(websocket, "receive_text"):
        return StarletteAdapter(websocket)

    # Check for websockets library
    if hasattr(websocket, "send") and hasattr(websocket, "recv"):
        return WebSocketsAdapter(websocket)

    raise ValueError(
        "Unsupported WebSocket type. "
        "Must be either websockets.ServerConnection "
        "or starlette.websockets.WebSocket. "
        f"Received: {type(websocket)}"
    )


def is_websocket_available() -> bool:
    """Check if any WebSocket library is available.

    Returns
    -------
    bool
        True if websockets or starlette is available, False otherwise.
    """
    return HAS_WS_LIB



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent specific extras module."""
# pylint: disable=too-few-public-methods

from dataclasses import dataclass

from ..enums import ContentOrder, ExportPosition
from ..result import ExportResult
from .base import BaseExtras


@dataclass
class ToolExtras(BaseExtras):
    """Extras for tool exporters."""

    function_content: str = ""
    registration_content: str = ""

    def add_function_content(self, content: str) -> None:
        """Add function definition content.

        Parameters
        ----------
        content : str
            The function content to add.
        """
        if content and content.strip():  # pragma: no branch
            if self.function_content:
                self.function_content += "\n\n" + content.rstrip()
            else:
                self.function_content = content.rstrip()
            while not self.function_content.endswith("\n\n"):
                self.function_content += "\n\n"

    def add_registration_content(self, content: str) -> None:
        """Add function registration content.

        Parameters
        ----------
        content : str
            The registration content to add.
        """
        if content and content.strip():  # pragma: no branch
            self.registration_content += "\n" + content.rstrip() + "\n"

    def has_specific_content(self) -> bool:
        """Check for tool specific content.

        Returns
        -------
        bool
            True if there's tool specific content.
        """
        return bool(
            self.function_content.strip() or self.registration_content.strip()
        )

    def _contribute_specific_content(self, result: ExportResult) -> None:
        """Contribute tool specific content.

        Parameters
        ----------
        result : ExportResult
            The export result to contribute to.
        """
        # Function definitions go in TOOLS section
        if self.function_content:
            result.add_content(
                self.function_content,
                ExportPosition.TOOLS,
                order=ContentOrder.PRE_CONTENT,
            )

        # Registration content goes after agents
        if self.registration_content:
            result.add_content(
                self.registration_content,
                ExportPosition.AGENTS,
                # After agents but before handoffs
                order=ContentOrder.CLEANUP,
            )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez captain agent model."""

from typing import Literal

from pydantic import Field
from typing_extensions import Annotated

from ..agent import WaldiezAgent
from .captain_agent_data import WaldiezCaptainAgentData


class WaldiezCaptainAgent(WaldiezAgent):
    """Captain agent.

    A `WaldiezAgent` with agent_type `captain` and
    captain agent's related config for the agent.
    Also see `WaldiezAgent`, `WaldiezCaptainData`, `WaldiezAgentData`

    Attributes
    ----------
    agent_type : Literal["captain"]
        The agent type: 'captain' for a captain agent
    data : WaldiezCaptainAgentData
        The captain agent's data.
    """

    agent_type: Annotated[  # pyright: ignore
        Literal["captain"],
        Field(
            default="captain",
            title="Agent type",
            description="The agent type: 'captain' for a captain agent",
            alias="agentType",
        ),
    ] = "captain"
    data: Annotated[  # pyright: ignore
        WaldiezCaptainAgentData,
        Field(
            title="Data",
            description="The captain agent's data",
            default_factory=WaldiezCaptainAgentData,  # pyright: ignore
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
"""Resoning agent extras module."""

from dataclasses import dataclass

from waldiez.exporting.core.result import ExportResult

from ...enums import AgentPosition, ExportPosition
from .standard_extras import StandardExtras


@dataclass
class ReasoningExtras(StandardExtras):
    """Extras for reasoning agents."""

    def _contribute_specific_content(self, result: ExportResult) -> None:
        """Contribute reasoning specific content to the export result.

        Parameters
        ----------
        result : ExportResult
            The export result to contribute to.
        """
        super()._contribute_specific_content(result)
        if self.extra_args:
            for arg in self.extra_args:
                result.add_instance_argument(
                    name=arg.name,
                    value=arg.value,
                    instance_id=arg.instance_id,
                    tabs=arg.tabs,
                )
        if self.extra_imports:
            for imp in self.extra_imports:
                result.add_import(imp.statement, imp.position)
        if self.before_agent.strip():
            result.add_content(
                self.before_agent,
                position=ExportPosition.AGENTS,
                agent_position=AgentPosition.BEFORE,
                agent_id=self.instance_id,
            )
        if self.after_agent.strip():
            result.add_content(
                self.after_agent,
                position=ExportPosition.AGENTS,
                agent_position=AgentPosition.AFTER,
                agent_id=self.instance_id,
            )

    def has_specific_content(self) -> bool:
        """Check for reasoning specific content.

        Returns
        -------
        bool
            True if there's reasoning specific content.
        """
        if not super().has_specific_content():
            return bool(
                self.extra_args
                or self.extra_imports
                or self.before_agent.strip()
                or self.after_agent.strip()
            )
        return True



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Factory function for creating a ModelsExporter instance."""

from pathlib import Path
from typing import Optional, Union

from waldiez.models import WaldiezAgent, WaldiezModel

from ..core import ExporterContext, get_default_exporter_context
from .exporter import ModelsExporter


def create_models_exporter(
    # Factory function for models exporter creation
    flow_name: str,
    agents: list[WaldiezAgent],
    agent_names: dict[str, str],
    models: list[WaldiezModel],
    model_names: dict[str, str],
    for_notebook: bool = False,
    cache_seed: Optional[int] = None,
    output_dir: Optional[Union[str, Path]] = None,
    context: Optional[ExporterContext] = None,
) -> ModelsExporter:
    """Create a models exporter.

    Parameters
    ----------
    flow_name : str
        The name of the flow.
    agents : list[WaldiezAgent]
        The agents that use models.
    agent_names : dict[str, str]
        Mapping of agent IDs to names.
    models : list[WaldiezModel]
        The models to export.
    model_names : dict[str, str]
        Mapping of model IDs to names.
    for_notebook : bool, optional
        Whether the export is for a notebook, by default False
    cache_seed : Optional[int], optional
        The cache seed if any, by default None
    output_dir : Optional[Union[str, Path]], optional
        Output directory for generated files, by default None
    context : Optional[ExporterContext], optional
        Exporter context with dependencies, by default None

    Returns
    -------
    ModelsExporter
        The created models exporter.
    """
    if context is None:
        context = get_default_exporter_context()
    return ModelsExporter(
        flow_name=flow_name,
        agents=agents,
        agent_names=agent_names,
        models=models,
        model_names=model_names,
        for_notebook=for_notebook,
        cache_seed=cache_seed,
        output_dir=output_dir,
        context=context,
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Flow export orchestrator."""

from typing import Any, Callable

from waldiez.models import Waldiez, WaldiezAgent

from ..agent import AgentExporter, create_agent_exporter
from ..chats import ChatsExporter, create_chats_exporter
from ..core import (
    AgentPosition,
    ContentOrder,
    ExporterContext,
    ExportPosition,
    ExportResult,
    ImportPosition,
)
from ..models import ModelsExporter, create_models_exporter
from ..tools import ToolsExporter, create_tools_exporter
from .merger import ContentMerger
from .utils import (
    generate_header,
    get_after_run_content,
    get_np_no_nep50_handle,
    get_sqlite_out,
    get_start_logging,
    get_stop_logging,
    get_the_imports_string,
)


# pylint: disable=no-self-use,too-many-instance-attributes
class ExportOrchestrator:
    """Coordinates the export process."""

    def __init__(
        self,
        waldiez: Waldiez,
        context: ExporterContext,
    ) -> None:
        """Initialize the export orchestrator.

        Parameters
        ----------
        waldiez : Waldiez
            The Waldiez instance containing the flow to export.
        context : ExporterContext
            The exporter context containing dependencies and configuration.
        """
        self.waldiez = waldiez
        self.context = context
        self.config = context.get_config()
        self._tools_exporter: ToolsExporter | None = None
        self._models_exporter: ModelsExporter | None = None
        self._chats_exporter: ChatsExporter | None = None
        self.logger = context.get_logger()
        self._initialize()

    def _initialize(self) -> None:
        """Initialize the orchestrator with necessary configurations."""
        unique_names = self.waldiez.flow.unique_names
        self.flow_name = unique_names["flow_name"]
        self.agents = unique_names["agents"]
        self.models = unique_names["models"]
        self.tools = unique_names["tools"]
        self.chats = unique_names["chats"]
        self.agent_names = unique_names["agent_names"]
        self.model_names = unique_names["model_names"]
        self.tool_names = unique_names["tool_names"]
        self.chat_names = unique_names["chat_names"]

    def orchestrate(self) -> ExportResult:
        """Orchestrate the export process.

        Returns
        -------
        ExportResult
            The result of the export process,
            containing the generated script and any additional metadata.
        """
        results: list[ExportResult] = []
        agent_arguments: dict[str, list[str]] = {}

        # 1. Tools first (needed by agents)
        if self.waldiez.tools:
            self.logger.info("Exporting tools ...")
            tools_result = self._get_tools_exporter().export()
            # Extract tool arguments for agents
            tool_arguments = self._extract_agent_arguments_from_result(
                tools_result
            )
            # Merge tool arguments into agent arguments
            self._merge_agent_arguments(
                source=tool_arguments,
                target=agent_arguments,
            )
            results.append(tools_result)
            self.logger.debug("Exported %s", tools_result)

        # 2. Models second (needed by agents)
        if self.waldiez.models:
            self.logger.info("Exporting models ...")
            models_result = self._get_models_exporter().export()
            # Extract model arguments for agents
            model_arguments = self._extract_agent_arguments_from_result(
                models_result
            )
            # Merge model arguments into agent arguments
            self._merge_agent_arguments(
                source=model_arguments,
                target=agent_arguments,
            )
            results.append(models_result)
            self.logger.debug("Exported %s", models_result)
        # 3. Chats third (agents might need agent chat registrations)
        # we always have at least one chat (already validated in Waldiez init)
        self.logger.info("Exporting chats ...")
        chats_result = self._get_chats_exporter().export()
        self.logger.debug("Exported %s", chats_result)
        # Extract chat arguments for agents
        chat_arguments = self._extract_agent_arguments_from_result(chats_result)
        # Merge chat arguments into agent arguments
        self._merge_agent_arguments(
            source=chat_arguments,
            target=agent_arguments,
        )
        results.append(chats_result)

        # 4. Agents last
        # we always have at least one agent (already validated in Waldiez init)
        agent_results = self._export_all_agents(agent_arguments)
        results.extend(agent_results)

        # 5. Merge everything
        merger = ContentMerger(self.context)
        merged_result = merger.merge_results(results)
        # Check for issues
        stats = merger.get_merge_statistics()
        if stats.conflicts_found:
            self.logger.info(
                "Resolved %d merge conflicts", len(stats.conflicts_found)
            )
        self.logger.debug("Merged result: %s", merged_result)
        return self._finalize_export(merged_result)

    def _finalize_export(self, merged_result: ExportResult) -> ExportResult:
        """Finalize the export result with additional content.

        Parameters
        ----------
        merged_result : ExportResult
            The merged export result containing all content.

        Returns
        -------
        ExportResult
            The finalized export result with additional content.
        """
        merged_result.add_content(
            generate_header(
                name=self.waldiez.name,
                description=self.waldiez.description,
                requirements=self.waldiez.requirements,
                tags=self.waldiez.tags,
                for_notebook=self.config.for_notebook,
            ),
            position=ExportPosition.TOP,  # befoe everything
            order=ContentOrder.EARLY_SETUP,
        )
        merged_result.add_content(
            get_np_no_nep50_handle(),
            position=ExportPosition.IMPORTS,  # after imports (need np)
            order=ContentOrder.CLEANUP,
        )
        merged_result.add_content(
            get_start_logging(
                is_async=self.waldiez.is_async,
                for_notebook=self.config.for_notebook,
            ),
            position=ExportPosition.IMPORTS,  # after imports, before models
            order=ContentOrder.CLEANUP.value + 1,  # after imports
            skip_strip=True,  # keep newlines
        )
        merged_result.add_content(
            get_sqlite_out(is_async=self.waldiez.is_async),
            position=ExportPosition.AGENTS,
            order=ContentOrder.LATE_CLEANUP.value + 1,  # after all agents
        )
        merged_result.add_content(
            get_stop_logging(is_async=self.waldiez.is_async),
            position=ExportPosition.AGENTS,
            order=ContentOrder.LATE_CLEANUP.value
            + 2,  # before def main (chats)
        )
        all_imports: list[tuple[str, ImportPosition]] = [
            (item.statement, item.position)
            for item in merged_result.get_sorted_imports()
        ]
        import_string = get_the_imports_string(
            all_imports=all_imports,
            is_async=self.waldiez.is_async,
        )
        if self.config.for_notebook:
            cell_ignore_comment = (
                "# pyright: "
                "reportUnusedImport=false,"
                "reportMissingTypeStubs=false"
                "\n"
            )
            import_string = cell_ignore_comment + import_string
        merged_result.add_content(
            import_string,
            position=ExportPosition.IMPORTS,  # imports section
            order=ContentOrder.EARLY_SETUP,  # top position
        )
        return merged_result

    def get_after_run_content(self) -> str:
        """Get the content to be executed after the main flow run.

        Returns
        -------
        str
            The content to be executed after the main flow run.
        """
        return get_after_run_content(
            waldiez=self.waldiez,
            agent_names=self.agent_names,
            tabs=1,
        )

    def _get_tools_exporter(self) -> ToolsExporter:
        """Get or create tools exporter."""
        if self._tools_exporter is None:
            self._tools_exporter = create_tools_exporter(
                flow_name=self.flow_name,
                agents=self.agents,
                agent_names=self.agent_names,
                tools=self.tools,
                tool_names=self.tool_names,
                output_dir=self.config.output_directory,
                context=self.context,
            )
        return self._tools_exporter

    def _get_models_exporter(self) -> ModelsExporter:
        """Get or create models exporter."""
        if self._models_exporter is None:
            self._models_exporter = create_models_exporter(
                flow_name=self.flow_name,
                agents=self.agents,
                agent_names=self.agent_names,
                models=self.models,
                model_names=self.model_names,
                for_notebook=self.config.for_notebook,
                cache_seed=self.config.cache_seed,
                output_dir=self.config.output_directory,
                context=self.context,
            )
        return self._models_exporter

    def _get_chats_exporter(self) -> ChatsExporter:
        """Get or create chats exporter."""
        if self._chats_exporter is None:
            self._chats_exporter = create_chats_exporter(
                waldiez=self.waldiez,
                all_agents=self.agents,
                agent_names=self.agent_names,
                models=self.models,
                model_names=self.model_names,
                tools=self.tools,
                tool_names=self.tool_names,
                all_chats=self.chats,
                main_chats=self.waldiez.initial_chats,
                chat_names=self.chat_names,
                output_dir=self.config.output_directory,
                context=self.context,
            )
        return self._chats_exporter

    # pylint: disable=no-self-use
    def _create_agent_arguments_resolver(
        self, exported_arguments: dict[str, Any]
    ) -> Callable[[WaldiezAgent], list[str]]:
        """Create an arguments resolver function for a specific agent.

        Parameters
        ----------
        exported_arguments : dict[str, Any]
            A dictionary containing exported arguments for agents,
            models, and tools.

        Returns
        -------
        Callable[[WaldiezAgent], list[str]]
            A function that takes a WaldiezAgent and returns a list of
            arguments to be used for that agent.
        """

        def arguments_resolver(target_agent: WaldiezAgent) -> list[str]:
            """Resolve arguments for the target agent.

            Parameters
            ----------
            target_agent : WaldiezAgent
                The agent for which to resolve arguments.

            Returns
            -------
            list[str]
                A list of arguments to be used for the agent.
            """
            # Extract arguments from exported_arguments
            return exported_arguments.get(target_agent.id, [])

        return arguments_resolver

    def _create_agent_exporter(
        self,
        agent: WaldiezAgent,
        agent_arguments: dict[str, list[str]],
    ) -> AgentExporter:
        """Create an exporter for a specific agent."""
        return create_agent_exporter(
            agent=agent,
            agent_names=self.agent_names,
            models=(self.models, self.model_names),
            chats=(self.chats, self.chat_names),
            initial_chats=self.waldiez.initial_chats,
            tool_names=self.tool_names,
            cache_seed=self.waldiez.cache_seed,
            is_async=self.waldiez.is_async,
            for_notebook=self.config.for_notebook,
            output_dir=self.config.output_directory,
            context=self.context,
            group_chat_members=self.waldiez.get_group_chat_members(agent),
            arguments_resolver=self._create_agent_arguments_resolver(
                exported_arguments=agent_arguments,
            ),
        )

    def _extract_agent_arguments_from_result(
        self, result: ExportResult
    ) -> dict[str, list[str]]:
        """Extract agent-specific arguments from an export result."""
        agent_args: dict[str, list[str]] = {}

        for content in result.positioned_content:
            # Look for content positioned as agent arguments
            if (
                content.position == ExportPosition.AGENTS
                and content.agent_position == AgentPosition.AS_ARGUMENT
                and content.agent_id
            ):
                if content.agent_id not in agent_args:
                    agent_args[content.agent_id] = []

                # The content itself is the argument string
                strip_content = content.content.strip()
                if strip_content:
                    # Add the content as an argument for the agent
                    agent_args[content.agent_id].append(f"    {strip_content}")

        return agent_args

    def _merge_agent_arguments(
        self, target: dict[str, list[str]], source: dict[str, list[str]]
    ) -> None:
        """Merge agent arguments from source into target."""
        for agent_id, args in source.items():
            if agent_id not in target:
                target[agent_id] = []
            target[agent_id].extend(args)

    def _gather_agent_arguments(self) -> dict[str, Any]:
        """Gather exported arguments from all agents, models, and tools.

        Returns
        -------
        dict[str, Any]
            A dictionary containing exported arguments for agents,
            models, and tools.
        """
        exported_arguments: dict[str, Any] = {}
        return exported_arguments

    def _export_all_agents(
        self, exported_arguments: dict[str, Any]
    ) -> list[ExportResult]:
        """Export all agents in dependency order.

        Parameters
        ----------
        exported_arguments : dict[str, Any]
            A dictionary containing exported arguments for agents,
            models, and tools.

        Returns
        -------
        list[ExportResult]
            A list of export results for each agent.
        """
        results: list[ExportResult] = []
        for agent in self.waldiez.agents:
            agent_exporter = self._create_agent_exporter(
                agent, exported_arguments
            )
            agent_result = agent_exporter.export()
            results.append(agent_result)
        return results



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Chats exporter."""

from typing import Any, Optional

from waldiez.models import (
    WaldiezAgent,
    WaldiezAgentConnection,
    WaldiezChat,
    WaldiezGroupManager,
)

from ..core import ChatExtras, Exporter, ExporterContext
from .processor import ChatsProcessor


class ChatsExporter(Exporter[ChatExtras]):
    """Chats exporter with structured extras."""

    def __init__(
        self,
        all_agents: list[WaldiezAgent],
        agent_names: dict[str, str],
        all_chats: list[WaldiezChat],
        chat_names: dict[str, str],
        main_chats: list[WaldiezAgentConnection],
        root_group_manager: Optional[WaldiezGroupManager],
        context: Optional[ExporterContext] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the chats exporter.

        Parameters
        ----------
        all_agents : list[WaldiezAgent]
            All agents involved in the chats.
        agent_names : dict[str, str]
            Mapping of agent IDs to their names.
        all_chats : list[WaldiezChat]
            All chats to be exported.
        chat_names : dict[str, str]
            Mapping of chat IDs to their names.
        main_chats : list[WaldiezAgentConnection]
            Main chats that are connections between agents.
        root_group_manager : Optional[WaldiezGroupManager]
            The root group manager for managing chat groups, if any.
        context : Optional[ExporterContext], optional
            Exporter context with dependencies, by default None
        **kwargs : Any
            Additional keyword arguments for the exporter.
        """
        super().__init__(context, **kwargs)

        self.all_agents = all_agents
        self.agent_names = agent_names
        self.all_chats = all_chats
        self.chat_names = chat_names
        self.main_chats = main_chats
        self.root_group_manager = root_group_manager
        config = self.context.get_config()
        self.for_notebook = config.for_notebook
        self.is_async = config.is_async
        self.cache_seed = config.cache_seed
        # Initialize extras with processed chat content
        self._extras = self._create_chat_extras()

    @property
    def extras(self) -> ChatExtras:
        """Get the chat extras."""
        return self._extras

    # pylint: disable=no-self-use
    def _create_chat_extras(self) -> ChatExtras:
        """Create and populate chat extras."""
        extras = ChatExtras("chats")
        processor = ChatsProcessor(
            all_agents=self.all_agents,
            agent_names=self.agent_names,
            all_chats=self.all_chats,
            chat_names=self.chat_names,
            main_chats=self.main_chats,
            root_group_manager=self.root_group_manager,
            for_notebook=self.for_notebook,
            is_async=self.is_async,
            cache_seed=self.cache_seed,
            serializer=self.context.get_serializer(),
            extras=extras,
        )
        processor.process()
        if extras.chat_registration:
            extras.append_after_all_agents(extras.chat_registration)
        return extras

    def generate_main_content(self) -> Optional[str]:
        """Generate the main content of the export."""
        return None



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Agent Nested Chat."""

from pydantic import Field
from typing_extensions import Annotated

from ...common import (
    WaldiezBase,
    WaldiezDefaultCondition,
    WaldiezHandoffCondition,
    WaldiezTransitionAvailability,
)


class WaldiezAgentNestedChatMessage(WaldiezBase):
    """Waldiez Agent nested chat message.

    A reference to a chat's message or reply in a nested chat

    Attributes
    ----------
    id : str
        The id of the chat.
    is_reply : bool
        Whether to use the reply in the chat or not.
    """

    id: Annotated[
        str, Field(..., title="ID", description="The id of the chat.")
    ]
    is_reply: Annotated[
        bool,
        Field(
            False,
            title="Is reply",
            description="Whether to use the reply in the chat or not.",
            alias="isReply",
        ),
    ]


class WaldiezAgentNestedChat(WaldiezBase):
    """Waldiez Agent Nested Chat.

    Attributes
    ----------
    triggered_by : list[str]
        A list of agent ids that trigger the nested chat.
    messages : list[WaldiezAgentNestedChatMessage]
        The list of messages (chat ids and 'is_reply'z)
        to include the in the nested chat registration.
    order : int
        The order of the nested chat (if used as a handoff).
        Defaults to 0.
    """

    triggered_by: Annotated[
        list[str],
        Field(
            title="Triggered By",
            description=("A list of agent ids that trigger the nested chat."),
            alias="triggeredBy",
            default_factory=list,
        ),
    ]
    messages: Annotated[
        list[WaldiezAgentNestedChatMessage],
        Field(
            title="Messages",
            description=(
                "The list of messages (chat ids and 'is_reply'z)"
                "to include the in the nested chat registration."
            ),
            default_factory=list,
        ),
    ]
    condition: Annotated[
        WaldiezHandoffCondition,
        Field(
            default_factory=WaldiezDefaultCondition.create,
            title="Condition",
            description=(
                "The condition to use for the nested chat handoff. "
                "If not provided, the nested chat will always be available."
            ),
        ),
    ]
    available: Annotated[
        WaldiezTransitionAvailability,
        Field(
            default_factory=WaldiezTransitionAvailability,
            title="Available",
            description=(
                "The availability of the nested chat. "
                "If not provided, the nested chat will always be available."
            ),
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Generate a unique ID."""

import uuid
from datetime import datetime, timezone


def get_id() -> str:
    """Generate a unique ID.

    Returns
    -------
    str
        The unique ID.
    """
    now_td = datetime.now(timezone.utc)
    now_str = now_td.strftime("%Y%m%d%H%M%S%f")
    return f"{now_str}-{uuid.uuid4().hex}"
